/* coordgeom.c
some coordinate geometry functions and ... er ... stuff
Prim Andras 2002
*/

#include <math.h>
#include <stdlib.h>
#include "coordgeom.h"
#include "myio.h"


vector3f nullvect = {0.0, 0.0, 0.0};

char whatpointindex[8][8] = {
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1,-1,-1,-1,-1,-1,
-1,-1,-1, 2, 1, 2, 1,-1,
-1,-1, 2,-1, 0, 2,-1, 0,
-1,-1, 1, 0,-1,-1, 1, 0,
-1,-1, 2, 2,-1, 2,-1,-1,
-1,-1, 1,-1, 1,-1, 1,-1,
-1,-1,-1, 0, 0,-1,-1, 0
};

char cliptype[8][8] = {
 inner, inner, inner, inner, inner, inner, inner, inner,
 inner, inner, inner, inner, inner, inner, inner, inner,
 inner, inner,  none,    AB,    AC,    AB,    AC, inner,
 inner, inner,    AB,  none,    BC,    AB, inner,    BC,
 inner, inner,    AC,    BC,  none, inner,    AC,    BC,
 inner, inner,    AB,    AB, inner,    AB, inner, inner,
 inner, inner,    AC, inner,    AC, inner,    AC, inner,
 inner, inner, inner,    BC,    BC, inner, inner,    BC
};

char needneighbour[8][8] = {
  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,N01,N02,N01,N02,  0,
  0,  0,N01,  0,N12,N01,  0,N12,
  0,  0,N01,N12,  0,  0,N02,N12,
  0,  0,N01,N01,  0,N01,  0,  0,
  0,  0,N02,  0,N02,  0,N02,  0,
  0,  0,  0,N12,N12,  0,  0,N12
};


/*  puts the normal vector of triangle 'tri' into 'normal'
*/
void GetNormal3f(triangle3f tri, vector3f normal)
{
  float a, b, c, d, e, f, x, y, z, i, tmp;
  a = tri[1][0] - tri[0][0]; b = tri[1][1] - tri[0][1]; c = tri[1][2] -
 tri[0][2];
  d = tri[2][0] - tri[0][0]; e = tri[2][1] - tri[0][1]; f = tri[2][2] -
 tri[0][2];

  z = e*a - d*b;
  y = c*d - f*a;
  x = f*b - c*e;

    // normalize the vector
  i = x*x + y*y + z*z;
  i = sqrt(i);
  normal[0] = x / i; normal[1] = y / i; normal[2] = z / i;
}


/* puts the endpoints of the section of triangles 'ta' 'tb' into 'va' 'vb'
 if the triangles does not intersect, 'va' = 'vb' = (0,0,0)
*/
void GetSection3f (triangle3f ta, triangle3f tb, vector3f va, vector3f vb)
{
// body removed at 2002-Marc-19 for having clearer source
  return;
}



/* roughly the same as GetSection3f, but receives the normals of the triangles
too, and the output is in sectline format (but does not alter the trindex
members of the sectline structures)
if no intersection found, sla->endtype[1] = none and slb->endtype[1] = none*/
void GetSectionn3f (triangle3f ta, vector3f norma, triangle3f tb,
vector3f normb, sectline *sla, sectline *slb)
{
  vertex3f sa[2], sb[2];
  tendtype eta[2], etb[2];
  float consta, constb, theta, theta2, thetainv, n, sz, ftmp;
  int i, j, ind, needb0, needb1;
  triangle3f triangletmp;
//debfuncbegin("GetSectionn3f\n");

    /* You know, the normal is quite useful to count the plane's equation.
    equation of plane of triangle ta:
    norma[0]*x + norma[1]*y + norma[2]*z = consta;
    */
  consta = norma[0]*ta[0][0] + norma[1]*ta[0][1] + norma[2]*ta[0][2];
  constb = normb[0]*tb[0][0] + normb[1]*tb[0][1] + normb[2]*tb[0][2];

    /* this indicates, that the triangles does not intersect
    it will be altered, if intersection found */
  sla->endtype[1] = none;
  slb->endtype[1] = none;

    /* Now, count where the plane of ta intersects triangle tb and vica versa.
       first: plane of ta and triangle tb, check every side of tb, whether it is
       intersected by the plane */
  for(i = 0, j = 1, ind = 0; (i < 3) && (ind < 2);
  j = ((j < 2) ? j+1 : 0), i++)
  {
    sz = norma[0]*tb[j][0] + norma[1]*tb[j][1] + norma[2]*tb[j][2];
      /* if tb[j] is on the plane of ta */
    if (nullf(sz - consta))
    {
      memcpy(&sb[ind], &tb[j], sizeof(vertex3f));
      etb[ind] = (j == 0 ? A : (j == 1 ? B : C));
      ind ++;
      continue;
    }

    n = norma[0]*tb[i][0] + norma[1]*tb[i][1] + norma[2]*tb[i][2] - sz;
      /* n ~ 0 if tb is near paralell to the plane */
    if (notnullf2(n))
    {
        /* equation of line AB: P = theta*A + (1-theta)*B
          P is between A and B if 0 < theta < 1 */
      theta = (consta - sz) / n;
//printf("theta %f ", theta);
        /* if the intersection point is on the triangle's side
        ( and by no way on a corner) */
      if (ltf2(0, theta) && ltf2(theta, 1))
      {
        thetainv = 1 - theta;
        sb[ind][0] = tb[i][0]*theta + tb[j][0]*thetainv;
        sb[ind][1] = tb[i][1]*theta + tb[j][1]*thetainv;
        sb[ind][2] = tb[i][2]*theta + tb[j][2]*thetainv;

        etb[ind] = (i == 0 ? AB : (i == 1 ? BC : AC));
        ind++;
      }
    }
  }
    /* less than 2 intersection point (that is probably zero) means no real
     intersection */
  if (ind < 2)
  {
    return;
  }
//debinfo("GetSectionn3f 1\n");

    /* plane of tb and triangle ta */
  for(i = 0, j = 1, ind = 0; (i < 3) && (ind < 2);
   j = ((j < 2) ? j+1 : 0), i++)
  {
    sz = normb[0]*ta[j][0] + normb[1]*ta[j][1] + normb[2]*ta[j][2];
      /* if ta[j] is on the plane of tb */
    if (nullf(sz - constb))
    {
      memcpy(&sa[ind], &ta[j], sizeof(vertex3f));
      eta[ind] = (j == 0 ? A : (j == 1 ? B : C));
      ind ++;
      continue;
    }

    n = normb[0]*ta[i][0] + normb[1]*ta[i][1] + normb[2]*ta[i][2] - sz;
      /* n ~ 0 if tb is near paralell to the plane */
    if (notnullf2(n))
    {
        /* equation of line AB: P = theta*A + (1-theta)*B
          P is between A and B if 0 < theta < 1 */
      theta = (constb - sz) / n;
        /* if the intersection point is on the triangle's side */
      if ( ltf2(0, theta) && ltf2(theta, 1) )
      {
        thetainv = 1 - theta;
        sa[ind][0] = ta[i][0]*theta + ta[j][0]*thetainv;
        sa[ind][1] = ta[i][1]*theta + ta[j][1]*thetainv;
        sa[ind][2] = ta[i][2]*theta + ta[j][2]*thetainv;

        eta[ind] = (i == 0 ? AB : (i == 1 ? BC : AC));
        ind++;
      }
    }
  }
    /* less than 2 intersection point (that is probably zero) means no real
     intersection */
  if (ind < 2)
  {
    sla->endtype[1] = none;
    return;
  }
//debinfo("GetSectionn3f 2\n");

    /* now we have sa[2]: the endpoints of the intersection of triangle ta and
     the plane of triangle tb; and sb[2]. We need their minimum */
  for(i = 0, n = 0, ind = -1; i < 3; i++)
  {
    sz = absf(sa[0][i] - sa[1][i]);
    if (ltf2(n,sz))
    {
      ind = i;
      n = sz;
    }
  }
    /* ind < 0 if sa[2] contains two identical points */
  if (ind < 0)
  {
//  printf("identical points (%f %f %f) (%f %f %f)\n",
//sa[1][0],sa[1][1],sa[1][2],   sa[0][0],sa[0][1],sa[0][2]);
    sla->endtype[1] = none;
    slb->endtype[1] = none;
    return;
  }

  n = sa[0][ind] - sa[1][ind];
  theta = (sb[0][ind] - sa[1][ind]) / n;
  if (ltf2(0, theta) && ltf2(theta, 1))
    needb0 = 1;
  else
    needb0 = 0;

  theta2 = (sb[1][ind] - sa[1][ind]) / n;
  if (ltf2(0 ,theta2) && ltf2(theta2, 1))
    needb1 = 1;
  else
    needb1 = 0;

  if (needb0 && needb1)
  {
//sla->endtype[1] = none; return;

       /* check if an endpoint is the same at both triangles */
    if (nullf(1-theta))
      sla->endtype[0] = eta[0];
    else if (nullf(theta))
      sla->endtype[0] = eta[1];
    else
      sla->endtype[0] = cliptype[eta[0]][eta[1]];

    if (nullf(1-theta2))
      sla->endtype[1] = eta[0];
    else if (nullf(theta2))
      sla->endtype[1] = eta[1];
    else
      sla->endtype[1] = cliptype[eta[0]][eta[1]];

    memcpy(sla->endpoint, sb, 2*sizeof(vertex3f));
    memcpy(slb->endpoint, sb, 2*sizeof(vertex3f));
    memcpy(slb->endtype, etb, 2*sizeof(tendtype));
  }
  else if (!(needb0 || needb1))
  {
//sla->endtype[1] = none; return;
      /* sa[2] sb[2] are on the same line but are disjoint*/
    if( (ltef(theta, 0) && ltef(theta2, 0)) ||
      (ltef(1, theta) && ltef(1, theta2)) )
    {
      sla->endtype[1] = none;
      slb->endtype[1] = none;
      return;
    }

    if (nullf(1-theta))
      slb->endtype[0] = etb[0];
    else if (nullf(1-theta2))
      slb->endtype[0] = etb[1];
    else
      slb->endtype[0] = cliptype[etb[0]][etb[1]];

    if (nullf(theta))
      slb->endtype[1] = etb[0];
    else if (nullf(theta2))
      slb->endtype[1] = etb[1];
    else
      slb->endtype[1] = cliptype[etb[0]][etb[1]];

    memcpy(sla->endpoint, sa, 2*sizeof(vertex3f));
    memcpy(sla->endtype, eta, 2*sizeof(tendtype));
    memcpy(slb->endpoint, sa, 2*sizeof(vertex3f));
  }
  else if (needb0)
  {
//sla->endtype[1] = none; return;
    memcpy(sla->endpoint[0], sb[0], sizeof(vertex3f));
    slb->endtype[0] = etb[0];

    if (ltef2(theta2, 0) )
    {
      memcpy(sla->endpoint[1], sa[1], sizeof(vertex3f));
      sla->endtype[1] = eta[1];

      if (nullf(1-theta))
        sla->endtype[0] = eta[0];
      else
        sla->endtype[0] = cliptype[eta[0]][eta[1]];

      if (nullf(theta2))
        slb->endtype[1] = etb[1];
      else
        slb->endtype[1] = cliptype[etb[0]][etb[1]];
    }
    else
    {
      memcpy(sla->endpoint[1], sa[0], sizeof(vertex3f));
      sla->endtype[1] = eta[0];

      if (nullf(theta))
        sla->endtype[0] = eta[1];
      else
        sla->endtype[0] = cliptype[eta[0]][eta[1]];

      if (nullf(1 - theta2))
        slb->endtype[1] = etb[1];
      else
        slb->endtype[1] = cliptype[etb[0]][etb[1]];
    }
    memcpy(slb->endpoint, sla->endpoint, 2*sizeof(vertex3f));
  }
  else /* if (needb1) */
  {
//sla->endtype[1] = none; return;
    memcpy(sla->endpoint[1], sb[1], sizeof(vertex3f));
    slb->endtype[1] = etb[1];

    if (ltef2(theta,0))
    {
      memcpy(sla->endpoint[0], sa[1], sizeof(vertex3f));
      sla->endtype[0] = eta[1];

      if (nullf(1-theta2))
        sla->endtype[1] = eta[0];
      else
        sla->endtype[1] = cliptype[eta[0]][eta[1]];

      if (nullf(theta))
        slb->endtype[0] = etb[0];
      else
        slb->endtype[0] = cliptype[etb[0]][etb[1]];
    }
    else
    {
      memcpy(sla->endpoint[0], sa[0], sizeof(vertex3f));
      sla->endtype[0] = eta[0];

      if (nullf(theta2))
        sla->endtype[1] = eta[1];
      else
        sla->endtype[1] = cliptype[eta[0]][eta[1]];

      if (nullf(1-theta))
        slb->endtype[0] = etb[0];
      else
        slb->endtype[0] = cliptype[etb[0]][etb[1]];
    }

    memcpy(slb->endpoint, sla->endpoint, 2*sizeof(vertex3f));
  }
/*
printf("theta %f, theta2 %f, eta[0] %d, eta[1] %d, etb[0] %d, etb[1] %d\n",
  theta, theta2, eta[0], eta[1], etb[0], etb[1] );
printf("endtypes: sla[0] %d, sla[1] %d, slb[0] %d, slb[1] %d\n",
sla->endtype[0], sla->endtype[1], slb->endtype[0], slb->endtype[1] );

if ((sla->endtype[0] > 1) && (sla->endtype[1] > 1))
  printf("a");
if ((slb->endtype[0] > 1) && (slb->endtype[1] > 1))
  printf("b");
*/

  memcpy(triangletmp[0], sa[0], sizeof(vertex3f));
  memcpy(triangletmp[1], sa[1], sizeof(vertex3f));
  memcpy(triangletmp[2], sa[0], sizeof(vertex3f));
  vertexadd3f(triangletmp[2], norma);
  GetNormal3f(triangletmp, sla->normal);
  ftmp = sla->normal[0]*normb[0] +
    sla->normal[1]*normb[1] +
    sla->normal[2]*normb[2];
  if(ltf2(ftmp,0))
  {
    memcpy(triangletmp[1], sa[0], sizeof(vertex3f));
    memcpy(triangletmp[0], sa[1], sizeof(vertex3f));
    GetNormal3f(triangletmp, sla->normal);
  }

  memcpy(triangletmp[0], sb[0], sizeof(vertex3f));
  memcpy(triangletmp[1], sb[1], sizeof(vertex3f));
  memcpy(triangletmp[2], sb[0], sizeof(vertex3f));
  vertexadd3f(triangletmp[2], normb);
  GetNormal3f(triangletmp, slb->normal);
  ftmp = slb->normal[0]*norma[0] +
    slb->normal[1]*norma[1] +
    slb->normal[2]*norma[2];
  if(ltf2(ftmp,0))
  {
    memcpy(triangletmp[1], sb[0], sizeof(vertex3f));
    memcpy(triangletmp[0], sb[1], sizeof(vertex3f));
    GetNormal3f(triangletmp, slb->normal);
  }
//debfuncend("GetSectionn3f end\n");

}


int GetParalellSection3f(triangle3f ta, vector3f na, triangle3f tb, vector3f nb,
sectline *sectarray)
{
  return 0;
} // GetParalellSection3f end

#ifdef COORDTEST
#include <GL/glut.h>
#endif


int ClipTriangles(mesh2 *pmeshin, mesh2 *pmeshout, sectline *pline, int linenum,
  int flipnormals)
{
  mtriangle *ptriangle;
  triangle3f triangletmp;
  vector3f normaltmp;
  int i, j, trindex;

    // resets all flag
//  AllTriangleFlagAndMask(pmeshin, 0x00);
/*
glEnable(GL_COLOR_LOGIC_OP);
glLogicOp(GL_INVERT);
glColor3f(0.5, 0.5, 0.5);
glBegin(GL_LINES);
  for (i = 0; i < linenum; i++)
  {
    glNormal3fv(pline[i].normal);
    glVertex3fv(pline[i].endpoint[0]);
    glVertex3fv(pline[i].endpoint[1]);
  }
glEnd();
glDisable(GL_COLOR_LOGIC_OP);
  return 0;
*/
  for (i = 0; i < linenum; i = j)
  {
    for (j = i, trindex = pline[j].trindex; (j < linenum) &&
      (trindex == pline[j].trindex); j++)
      pline[j].flag = 0;

    ClipTriangle (pmeshin, pmeshout, trindex, &pline[i], j-i, flipnormals);
    SetTriangleFlag (pmeshin, trindex, TOUCHED);
  }

  for (i = 0; i < linenum; i = j)
  {
    ptriangle = getstoreitemaddress(pmeshin->trianglestore, pline[i].trindex);
    AddNeededNeighbours(ptriangle , pmeshin, pmeshout);
    for (j = i, trindex = pline[j].trindex; (j < linenum) &&
      (trindex == pline[j].trindex); j++);
  }

  return 0;
}

/* clips triangle of 'pmeshin' with index 'trindex' according to clipping
  lines in 'pline'
  adds new triangles to 'pmeshout' */
int ClipTriangle (mesh2 *pmeshin, mesh2 *pmeshout, int trindex, sectline *pline,
int linenum, int flipnormals)
{
  int lineused = 0, i, dummy, noline, index;
  stack linestack;
  triangle3f triangle;
  sectline sectlinetmp, *psectlinetmp;
  vertex3f vertexbegin, vertextmp;
  vector3f normal;
  mtriangle *ptriangle;

debfuncbegin("ClipTriangle\n");
  initstack(&linestack);

  for (i = 0; i < linenum; i++)
    pline[i].flag = 0;
  ptriangle = getstoreitemaddress(pmeshin->trianglestore,trindex);
  GetMeshTriangleAndNormal3f(pmeshin, trindex, triangle, normal);
  if (flipnormals)
  {
    normal[0] *= -1;
    normal[1] *= -1;
    normal[2] *= -1;
  }

#ifdef COORDTEST
/*glDisable(GL_LIGHTING);
glBegin(GL_LINES);
  for (i = 0; i < linenum; i++)
  {
    DrawLineAndNormalGL(&pline[i]);
  }
glEnd(); */
#endif

  while (lineused < linenum)
  {
    i = 0;
      // first two line of next sectline loop
    noline = selectnextline(pline, linenum, ptriangle, NULL, &sectlinetmp,
       pmeshin->transformmatrix);
    if (noline)
      return 1;
    memcpy (vertexbegin, sectlinetmp.endpoint[0], sizeof(vertex3f));
    psectlinetmp = push(&linestack, &sectlinetmp, sizeof(sectline));

/*
glVertex3fv(sectlinetmp.endpoint[0]);
glVertex3fv(sectlinetmp.endpoint[1]);
*/

    i ++;
    if ( !(sectlinetmp.flag & SIDE))
      lineused++;
      ptriangle->flag |=
        needneighbour[sectlinetmp.endtype[0]][sectlinetmp.endtype[1]];


    while ( (i < CLIPSAFENUM) && !noline &&
      !nearequal3f(psectlinetmp->endpoint[1], vertexbegin))
    {

      noline = selectnextline(pline, linenum, ptriangle, psectlinetmp,
         &sectlinetmp, pmeshin->transformmatrix);
      if (!noline)
      {
/*
glVertex3fv(sectlinetmp.endpoint[0]);
glVertex3fv(sectlinetmp.endpoint[1]);
*/
        psectlinetmp = push(&linestack, &sectlinetmp, sizeof(sectline));
        i ++;
        if ( !(sectlinetmp.flag & SIDE))
          lineused++;
        ptriangle->flag |=
          needneighbour[sectlinetmp.endtype[0]][sectlinetmp.endtype[1]];
      }
    }

    if (i > 2)
    {
      psectlinetmp = getstackunsafe(&linestack, &dummy);


      AddPolygonToMesh(pmeshout, psectlinetmp, i, normal);
/*      for(dummy = 0; dummy < i; dummy++)
        printf("(%d %d) ", psectlinetmp[dummy].endtype[0],
psectlinetmp[dummy].endtype[1]);
   puts("");*/

    }
/*
if(i == CLIPSAFENUM)
{
  printf("\n{(%f %f %f) (%f %f %f) (%f %f %f)}\n",
    triangle[0][0],triangle[0][1],triangle[0][2],
    triangle[1][0],triangle[1][1],triangle[1][2],
    triangle[2][0],triangle[2][1],triangle[2][2]);

  printf ("linenum: %d lineused %d\n", linenum, lineused);
  for (i = 0; i < linenum; i++)
  {
      printf ("type: %d %d",pline[i].endtype[0], pline[i].endtype[1]);
      printf ("[(%f %f %f)", pline[i].endpoint[0][0],
       pline[i].endpoint[0][1], pline[i].endpoint[0][2]);
      printf (" (%f %f %f)]\n", pline[i].endpoint[1][0],
       pline[i].endpoint[1][1], pline[i].endpoint[1][2]);
//      DrawLineAndNormalGL(&pline[i]);
  }

  for(i = 0; i < CLIPSAFENUM; i++)
    printf("(%d %d) ", psectlinetmp[i].endtype[0],
psectlinetmp[i].endtype[1]);
}
*/

    zerostack(linestack);
  }
  freestack(&linestack);
debfuncend("ClipTriangle end\n");
  return 0;
}

int selectnextline(sectline *pline, int linenum, mtriangle *ptriangle, sectline
*ppreviousline, sectline *plineout, matrix4x4f transformmatrix)
{
  int i, j, besti, bestendindex;
  triangle3f triangletmp, triangletmp2;
  vector3f vectortmp;
  vertex3f possible1, possible2, *sa, *sb, vertextmp;
  float theta, besttheta, n;
  tendtype innerendtype = none, possible1type, possible2type;
debfuncbegin("selectnextline\n");


  if(NULL == ppreviousline)
  {
    for (i = 0; (i < linenum) && (pline[i].flag & TOUCHED); i++);
    if(i == linenum)
    {
      plineout->endtype[0] = none;
      return 1;
    }
    else
    {
      memcpy(plineout, &pline[i], sizeof(sectline));
      plineout->flag = 0;
      pline[i].flag |= TOUCHED | FIRSTSECT;

//DrawLineAndNormalGL(&pline[i]);

    }

debfuncend("selectnextline end a\n");
    return 0;
  }

  plineout->flag = 0;

//printf ("ppreviousline->endtype[1] %d\n", ppreviousline->endtype[1]);
  memcpy(plineout->endpoint[0],ppreviousline->endpoint[1],
    sizeof(vertex3f));
  plineout->endtype[0] = ppreviousline->endtype[1];

    // look, if anyone in pline is appropriate
  for(i = 0; i < linenum; i++)
  {
    if (!(pline[i].flag & TOUCHED))
    {
      if (nearequal3f(pline[i].endpoint[0], ppreviousline->endpoint[1]))
      {
        memcpy(plineout->endpoint[1],pline[i].endpoint[1],
          sizeof(vertex3f));
        plineout->endtype[1] = pline[i].endtype[1];
        memcpy (plineout->normal, pline[i].normal, sizeof(vector3f));
        // that's irrelevant
        // plineout->trindex = pline[i].trindex;
        pline[i].flag |= TOUCHED;

//DrawLineAndNormalGL(&pline[i]);

debfuncend("selectnextline end b\n");
        return 0;
      }
      else if(nearequal3f(pline[i].endpoint[1], ppreviousline->endpoint[1]))
      {
        memcpy(plineout->endpoint[1],pline[i].endpoint[0],
          sizeof(vertex3f));
        plineout->endtype[1] = pline[i].endtype[0];
        memcpy (plineout->normal, pline[i].normal, sizeof(vector3f));
        // taht's irrelevant
        // plineout->trindex = pline[i].trindex;
        pline[i].flag |= TOUCHED;

//DrawLineAndNormalGL(&pline[i]);

debfuncend("selectnextline end c\n");
        return 0;
      }
    }
  }

  // we reach this point, if noone in pline is appropriate
  // check whether a triangle side is good
  paOuterMatrixVertex3f(transformmatrix,
    ptriangle->angles[0]->coord, triangletmp2[0]);
  paOuterMatrixVertex3f(transformmatrix,
    ptriangle->angles[1]->coord, triangletmp2[1]);
  paOuterMatrixVertex3f(transformmatrix,
    ptriangle->angles[2]->coord, triangletmp2[2]);
  paOuterMatrixVector3f(transformmatrix,
    ptriangle->normal, vectortmp);



  switch(ppreviousline->endtype[1])
  {
  case A:
  case BC:
    memcpy(possible1, triangletmp2[1], sizeof(vertex3f));
    memcpy(possible2, triangletmp2[2], sizeof(vertex3f));
    possible1type = B;
    possible2type = C;
    break;
  case B:
  case AC:
    memcpy(possible1, triangletmp2[2], sizeof(vertex3f));
    memcpy(possible2, triangletmp2[0], sizeof(vertex3f));
    possible1type = C;
    possible2type = A;
    break;
  case C:
  case AB:
    memcpy(possible1, triangletmp2[0], sizeof(vertex3f));
    memcpy(possible2, triangletmp2[1], sizeof(vertex3f));
    possible1type = A;
    possible2type = B;
    break;
  default:
debfuncend("selectnextline end d\n");
    return 1;
  }
  if (goodpoint(ppreviousline, possible1))
    i = 1;
  else if (goodpoint(ppreviousline, possible2))
    i = 2;
    /* if neither possible1 nor possible2 is goodpoint, then
    they are on the line of ppreviousline */
  else if (distance3f(ppreviousline->endpoint[1], possible1) >
    distance3f(ppreviousline->endpoint[1], possible2))
    i = 1;
  else
    i = 2;

    /* calculate the normal of plineout, and put ptriangle's appropriate
 angle in plineout->endpoint[index]. It may be changed later */
  if (1 == i)
  {
    memcpy(plineout->endpoint[1],possible1, sizeof(vertex3f));
    plineout->endtype[1] = possible1type;
    switch(ppreviousline->endtype[1])
    {
    case A:
    case AB:
      innerendtype = AB;
      memcpy(triangletmp[0], triangletmp2[0], sizeof(vertex3f));
      memcpy(triangletmp[LEINDEX], triangletmp2[1], sizeof(vertex3f));
      break;
    case B:
    case BC:
      innerendtype = BC;
      memcpy(triangletmp[0], triangletmp2[1], sizeof(vertex3f));
      memcpy(triangletmp[LEINDEX], triangletmp2[2], sizeof(vertex3f));
      break;
    case C:
    case AC:
      innerendtype = AC;
      memcpy(triangletmp[0], triangletmp2[2], sizeof(vertex3f));
      memcpy(triangletmp[LEINDEX], triangletmp2[0], sizeof(vertex3f));
      break;
    }
    memcpy(triangletmp[LEINDEXN], vectortmp, sizeof(vector3f));
    vertexadd3f(triangletmp[LEINDEXN], triangletmp[0]);
    GetNormal3f(triangletmp, plineout->normal);
  }
  else
  {
    memcpy(plineout->endpoint[1],possible2, sizeof(vertex3f));
    plineout->endtype[1] = possible2type;
    switch(ppreviousline->endtype[1])
    {
    case B:
    case AB:
      innerendtype = AB;
      memcpy(triangletmp[0], triangletmp2[0], sizeof(vertex3f));
      memcpy(triangletmp[LEINDEX], triangletmp2[1], sizeof(vertex3f));
      break;
    case C:
    case BC:
      innerendtype = BC;
      memcpy(triangletmp[0], triangletmp2[1], sizeof(vertex3f));
      memcpy(triangletmp[LEINDEX], triangletmp2[2], sizeof(vertex3f));
      break;
    case A:
    case AC:
      innerendtype = AC;
      memcpy(triangletmp[0], triangletmp2[2], sizeof(vertex3f));
      memcpy(triangletmp[LEINDEX], triangletmp2[0], sizeof(vertex3f));
      break;
    }
    memcpy(triangletmp[LEINDEXN], vectortmp, sizeof(vector3f));
    vertexadd3f(triangletmp[LEINDEXN], triangletmp[0]);
    GetNormal3f(triangletmp, plineout->normal);
  }

  /* perhaps there is a sectline in pline, which one's endpoint is between the
  current endpoint and the recently found angle
    */
  sa = plineout->endpoint;
    // search a coordinate, which can be used
  for(j = 0; (j < 3) && (nullf(sa[1][j]-sa[0][j])); j++);
  n = sa[0][j] - sa[1][j];


  for(besttheta = 0, besti = -1, i = 0; i < linenum; i++)
    if (!(pline[i].flag & TOUCHED) || (pline[i].flag & FIRSTSECT))
    {
      if (pline[i].endtype[0] == innerendtype)
      {
        theta = (pline[i].endpoint[0][j] - sa[1][j]) / n;
        if (ltf2(theta, 1) && (theta > besttheta))
        {
          bestendindex = 0;
          besttheta = theta;
          besti = i;
        }
      }
      else if (pline[i].endtype[1] == innerendtype)
      {
        theta = (pline[i].endpoint[1][j] - sa[1][j]) / n;
        if (ltf2(theta, 1) && (theta > besttheta))
        {
          bestendindex = 1;
          besttheta = theta;
          besti = i;
        }
      }
    }

  if (besti >= 0)
  {
    memcpy(plineout->endpoint[1], pline[besti].endpoint[bestendindex],
      sizeof(vertex3f));
    plineout->endtype[1] = innerendtype;
  }

  plineout->flag |= SIDE;
//  DrawLineAndNormalGL(plineout);

debfuncend("selectnextline end e\n");

  return 0;
}


int AddNeededNeighbours(mtriangle *ptriangle, mesh2 *pmeshin, mesh2 *pmeshout)
{
  mtriangle *ptriangle2;
  triangle3f triangletmp;

  if (NULL == ptriangle)
    return 1;

  ptriangle->flag |= TOUCHED;

  if(ptriangle->flag & NEIGHBOUR01)
  {
    ptriangle2 = ptriangle->neighbours[0];
    if ((NULL != ptriangle2) && !(ptriangle2->flag & TOUCHED))
    {
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[0]->coord, triangletmp[0]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[1]->coord, triangletmp[1]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[2]->coord, triangletmp[2]);
      AddTriangleToMesh(pmeshout, triangletmp);
      AddAllNeighbours(ptriangle2, pmeshin, pmeshout);
    }
  }

  if(ptriangle->flag & NEIGHBOUR02)
  {
    ptriangle2 = ptriangle->neighbours[1];
    if ((NULL != ptriangle2) && !(ptriangle2->flag & TOUCHED))
    {
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[0]->coord, triangletmp[0]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[1]->coord, triangletmp[1]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[2]->coord, triangletmp[2]);
      AddTriangleToMesh(pmeshout, triangletmp);
      AddAllNeighbours(ptriangle2, pmeshin, pmeshout);
    }
  }

  if(ptriangle->flag & NEIGHBOUR12)
  {
    ptriangle2 = ptriangle->neighbours[2];
    if ((NULL != ptriangle2) && !(ptriangle2->flag & TOUCHED))
    {
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[0]->coord, triangletmp[0]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[1]->coord, triangletmp[1]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[2]->coord, triangletmp[2]);
      AddTriangleToMesh(pmeshout, triangletmp);
      AddAllNeighbours(ptriangle2, pmeshin, pmeshout);
    }
  }
  return 0;
}


int AddAllNeighbours(mtriangle *ptriangle, mesh2 *pmeshin, mesh2 *pmeshout)
{
  mtriangle *ptriangle2;
  triangle3f triangletmp;

  if (NULL == ptriangle)
    return 1;

  ptriangle->flag |= TOUCHED;

    ptriangle2 = ptriangle->neighbours[0];
    if ((NULL != ptriangle2) && !(ptriangle2->flag & TOUCHED))
    {
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[0]->coord, triangletmp[0]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[1]->coord, triangletmp[1]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[2]->coord, triangletmp[2]);
      AddTriangleToMesh(pmeshout, triangletmp);
      AddAllNeighbours(ptriangle2, pmeshin, pmeshout);
    }

    ptriangle2 = ptriangle->neighbours[1];
    if ((NULL != ptriangle2) && !(ptriangle2->flag & TOUCHED))
    {
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[0]->coord, triangletmp[0]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[1]->coord, triangletmp[1]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[2]->coord, triangletmp[2]);
      AddTriangleToMesh(pmeshout, triangletmp);
      AddAllNeighbours(ptriangle2, pmeshin, pmeshout);
    }

    ptriangle2 = ptriangle->neighbours[2];
    if ((NULL != ptriangle2) && !(ptriangle2->flag & TOUCHED))
    {
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[0]->coord, triangletmp[0]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[1]->coord, triangletmp[1]);
      paOuterMatrixVertex3f(pmeshin->transformmatrix,
        ptriangle2->angles[2]->coord, triangletmp[2]);
      AddTriangleToMesh(pmeshout, triangletmp);
      AddAllNeighbours(ptriangle2, pmeshin, pmeshout);
    }
  return 0;
}


int sectcompare (const void *a, const void *b)
{
  return ((sectline*)a)->trindex - ((sectline*)b)->trindex;
}

  /* computes the triangles of a union */
int Bool (mesh2 *pamesh, mesh2 *pbmesh, mesh2 *pnewmesh, int what)
{
  int ia, ja, ib, jb, i, jamax, jbmax, index;
  float la, lb;
    // I'll store section lines in them
  stack astack, bstack;
  sectline asect, bsect, *pline, sectarray[6];

  vertex3f vertextmp;
  triangle3f trianglea, triangleb;
  vector3f normala, normalb;
  char mask;

  mtriangle *ptriangle;

debfuncbegin("Bool\n");
  mask = ~(TOUCHED | N01 | N02 | N12);
  AllTriangleFlagAndMask(pamesh, mask);
  AllTriangleFlagAndMask(pbmesh, mask);

  initstack(&astack);
  initstack(&bstack);

  for(ja = 0, jamax = getstoreitemnum(*pamesh->trianglestore); ja < jamax; ja++)
    for(jb = 0, jbmax = getstoreitemnum(*pbmesh->trianglestore); jb < jbmax;
        jb++)
         {
          asect.trindex = ja;
          bsect.trindex = jb;
          GetMeshTriangleAndNormal3f (pamesh, ja, trianglea, normala);
          GetMeshTriangleAndNormal3f (pbmesh, jb, triangleb, normalb);
          if (paralell3f(normala, normalb))
          {
            if((UNION == what) || (INTERSECTION == what))
            {
              i = GetParalellSection3f(trianglea, normala, triangleb, normalb,
                sectarray);
              for (i--; i >=0; i--)
              {
                sectarray[i].trindex = ja;
                push(&astack, &sectarray[i], sizeof(sectline));
              }
            }
            //@
          }
          else
          {
            GetSectionn3f( trianglea, normala,
              triangleb, normalb,
              &asect, &bsect);
/*            la = distance3f(asect.normal, nullvect);
            lb = distance3f(bsect.normal, nullvect);
*/
            if (asect.endtype[1] != none)
            {
              if (notnullf(distance3f(asect.endpoint[0],asect.endpoint[1])))
              {
                SetTriangleFlag(pamesh, ja, TOUCHED);
                if ( (INTERSECTION == what) || (BMINUSA == what))
                {
                  asect.normal[0] *=-1;
                  asect.normal[1] *=-1;
                  asect.normal[2] *=-1;
                }

                index = whatpointindex[asect.endtype[0]][asect.endtype[1]];
                if ((0 > index) || (goodpoint(&asect, trianglea[index])))
                  push(&astack, &asect, sizeof(sectline));
              }

              if (notnullf(distance3f(bsect.endpoint[0],bsect.endpoint[1])))
              {
                SetTriangleFlag(pbmesh, jb, TOUCHED);
                if ( (INTERSECTION == what) || (AMINUSB == what))
                {
                  bsect.normal[0] *=-1;
                  bsect.normal[1] *=-1;
                  bsect.normal[2] *=-1;
                }

                index = whatpointindex[bsect.endtype[0]][bsect.endtype[1]];
                if ((0 > index) || (goodpoint(&bsect, triangleb[index])))
                {

                  push(&bstack, &bsect, sizeof(sectline));
                }
              }

            }
          }
        }
      // these are the section lines of pamesh[ia]
  pline = (sectline*) getstackunsafe(&astack, &ia);
  ia /= sizeof(sectline);

  if (BMINUSA == what)
    ClipTriangles(pamesh, pnewmesh, pline, ia, 1);
  else
    ClipTriangles(pamesh, pnewmesh, pline, ia, 0);

  freestack (&astack);


  pline = getstackunsafe(&bstack, &ib);
  ib /= sizeof(sectline);
  qsort(pline, ib, sizeof(sectline), sectcompare);

  if (AMINUSB == what)
    ClipTriangles(pbmesh, pnewmesh, pline, ib, 1);
  else
    ClipTriangles(pbmesh, pnewmesh, pline, ib, 0);

  freestack(&bstack);

  if ((ia == 0) && (ib == 0))
  {
    if( (UNION == what) || (AMINUSB == what))
    {
      jamax = GetMeshTriangleNum(pamesh);
      for(ja = 0; ja < jamax; ja++)
      {
        GetMeshTriangle3f(pamesh, ja, trianglea);
        AddTriangleToMesh(pnewmesh, trianglea);
      }
    }

    if( (UNION == what) || (BMINUSA == what))
    {
      jbmax = GetMeshTriangleNum(pbmesh);
      for(jb = 0; jb < jbmax; jb++)
      {
        GetMeshTriangle3f(pbmesh, jb, triangleb);
        AddTriangleToMesh(pnewmesh, triangleb);
      }
    }
  }
debfuncend("Bool end\n");

  return 0;
}


 /* returns true, if vectors a and b are (near) paralell */
int paralell3f (vector3f a, vector3f b)
{
  float rat;
  if (notnullf(b[0]))
  {
    rat = a[0]/b[0];
    if(nullf2(a[1] - rat*b[1]) && nullf2(a[2] - rat*b[2]))
      return 1;
  }
  else if (notnullf(b[1]))
  {
    rat = a[1]/b[1];
    if(nullf2(a[0] - rat*b[0]) && nullf2(a[2] - rat*b[2]))
      return 1;
  }
  else if (notnullf(b[2]))
  {
    rat = a[2]/b[2];
    if(nullf2(a[0] - rat*b[0]) && nullf2(a[1] - rat*b[1]))
      return 1;
  }

  return 0;
}


 /* returns true, if vectors a and b are (near) the same */
int nearequal3f (vector3f a, vector3f b)
{
  float ftmp, x, y, z;
  x = (a[0]-b[0]);
  y = (a[1]-b[1]);
  z = (a[2]-b[2]);
  ftmp = sqrt(x*x + y*y + z*z);
  return nullf(ftmp);
//  return (nullf(a[0]-b[0]) && nullf(a[1]-b[1]) && nullf(a[2]-b[2]));
}

int concave(sectline *a, sectline *b)
{
  vector3f vectortmp;
  float ftmp;

  vectortmp[0] = b->endpoint[1][0] - b->endpoint[0][0];
  vectortmp[1] = b->endpoint[1][1] - b->endpoint[0][1];
  vectortmp[2] = b->endpoint[1][2] - b->endpoint[0][2];
  ftmp = a->normal[0]*vectortmp[0] + a->normal[1]*vectortmp[1] +
      a->normal[2]*vectortmp[2];
  if (ltef2 (ftmp, 0))
    return 1;

  vectortmp[0] = a->endpoint[0][0] - a->endpoint[1][0];
  vectortmp[1] = a->endpoint[0][1] - a->endpoint[1][1];
  vectortmp[2] = a->endpoint[0][2] - a->endpoint[1][2];
  ftmp = b->normal[0]*vectortmp[0] + b->normal[1]*vectortmp[1] +
      b->normal[2]*vectortmp[2];
  if(ltef2(ftmp, 0))
    return 1;

  return 0;
}

float distance3f (vertex3f a, vertex3f b)
{
  float x, y, z;
  x = a[0] - b[0];
  y = a[1] - b[1];
  z = a[2] - b[2];

  return sqrt(x*x + y*y + z*z);
}


int samesidenormal(sectline *a, sectline *b)
{
  vector3f vectortmp;
  float one ,another;

  vectortmp[0] = b->endpoint[1][0] - b->endpoint[0][0];
  vectortmp[1] = b->endpoint[1][1] - b->endpoint[0][1];
  vectortmp[2] = b->endpoint[1][2] - b->endpoint[0][2];
  one = a->normal[0]*vectortmp[0] + a->normal[1]*vectortmp[1] +
      a->normal[2]*vectortmp[2];

  vectortmp[0] = a->endpoint[0][0] - a->endpoint[1][0];
  vectortmp[1] = a->endpoint[0][1] - a->endpoint[1][1];
  vectortmp[2] = a->endpoint[0][2] - a->endpoint[1][2];
  another = b->normal[0]*vectortmp[0] + b->normal[1]*vectortmp[1] +
      b->normal[2]*vectortmp[2];

  if((ltef2(one, 0) && ltef2(another, 0)) || (ltf2(0, one) && ltf2(0,another)))
    return 1;
  return 0;
}

int goodpoint(sectline *pline, vertex3f vertex)
{
  vector3f vectortmp;
  float ftmp;

  vectortmp[0] = vertex[0] - pline->endpoint[1][0];
  vectortmp[1] = vertex[1] - pline->endpoint[1][1];
  vectortmp[2] = vertex[2] - pline->endpoint[1][2];

  ftmp = pline->normal[0]*vectortmp[0] + pline->normal[1]*vectortmp[1] +
      pline->normal[2]*vectortmp[2];

  if (ltf2(0, ftmp))
      return 1;
  return 0;
}


int islinebetween(sectline *pline, int linenum, vector3f normal, int i, int j)
{
  float theta, a, b, c, ftmp;
  int index;
  sectline secttmp;
  triangle3f triangletmp;

  memcpy(secttmp.endpoint[0], pline[i].endpoint[0], sizeof(vertex3f));
  memcpy(secttmp.endpoint[1], pline[j].endpoint[1], sizeof(vertex3f));
    /* the normal of the third line of the triangle defined by
     pline[i] and pline[j] */
  memcpy(triangletmp[0], pline[i].endpoint[0], sizeof(vertex3f));
  memcpy(triangletmp[1], pline[j].endpoint[1], sizeof(vertex3f));
  memcpy(triangletmp[2], pline[i].endpoint[0], sizeof(vertex3f));
  vertexadd3f(triangletmp[2], normal);
  GetNormal3f(triangletmp, secttmp.normal);


  if(!goodpoint(&secttmp, pline[i].endpoint[1]))
  {
    memcpy(triangletmp[0], pline[j].endpoint[1], sizeof(vertex3f));
    memcpy(triangletmp[1], pline[i].endpoint[0], sizeof(vertex3f));
    GetNormal3f(triangletmp, secttmp.normal);
  }

  for (index = 0; index < linenum; index++)
  {
    if(!(pline[index].flag & TOUCHED) && (index != i) && (index != j))
    {
      if(goodpoint(&pline[i],pline[index].endpoint[0])
        && goodpoint(&pline[j],pline[index].endpoint[0])
        && goodpoint(&secttmp,pline[index].endpoint[0]))
      {
        return 1;
      }
      if(goodpoint(&pline[i],pline[index].endpoint[1])
        && goodpoint(&pline[j],pline[index].endpoint[1])
        && goodpoint(&secttmp,pline[index].endpoint[1]))
        return 1;
    }
  }
  return 0;
}


float howparalell (sectline *a, sectline *b)
{
  vertex3f vertextmp;
  float n;

  vertextmp[0] = b->endpoint[1][0] - b->endpoint[0][0];
  vertextmp[1] = b->endpoint[1][1] - b->endpoint[0][1];
  vertextmp[2] = b->endpoint[1][2] - b->endpoint[0][2];
  n = sqrt(vertextmp[0]*vertextmp[0] +
    vertextmp[1]*vertextmp[1] +
    vertextmp[2]*vertextmp[2]);
  vertexdiv3f(vertextmp, n);

  return (absf(a->normal[0] * vertextmp[0] +
    a->normal[1] * vertextmp[1] +
    a->normal[2] * vertextmp[2]));
}


#ifdef DEBUG
void printftriangle(triangle3f triangle)
{
  printf("{(%f %f %f) (%f %f %f) (%f %f %f)}\n",
    triangle[0][0],triangle[0][1],triangle[0][2],
    triangle[1][0],triangle[1][1],triangle[1][2],
    triangle[2][0],triangle[2][1],triangle[2][2]);
}

void printfvertex(vertex3f vertex)
{
  printf("[%f %f %f]\n", vertex[0], vertex[1], vertex[2]);
}
#endif

