/* coordgeom.h
some coordinate geometry functions and stuff
Prim Andras 2002
*/

#include "pastuff.h"
#include "pagl.h"
#include "mesh2.h"

#ifndef COORDGEM_H
#define COORDGEM_H

#define CLIPSAFENUM 20

#define absf(a) ((a) > 0.0 ? (a) : -(a))


#define UNION        1
#define INTERSECTION 2
#define AMINUSB      3
#define BMINUSA      4

#define LEINDEX  2
#define LEINDEXN 1




/*  puts the normal vector of triangle 'tri' into 'normal'
*/
void GetNormal3f (triangle3f tri, vector3f normal);

/* puts the endpoints of the section of triangles 'ta' 'tb' into 'va' 'vb'
 if the triangles does not intersect, 'va' = 'vb' = (0,0,0)
*/
void GetSection3f (triangle3f ta, triangle3f tb, vertex3f va, vertex3f vb);
/* roughly the same as GetSection3f, but receives the normals of the triangles
too, and the output is in sectline format (but does not alter the trindex
members of the sectline structures) */
void GetSectionn3f (triangle3f ta, vector3f na, triangle3f tb, vector3f nb,
 sectline *sla, sectline *slb);
/* puts the bounding polygon of intersection of paralell triangles 'ta' and 'tb'
 into 'sectarray', returns the number of lines in sectarray (maximum 6)
*/
int GetParalellSection3f(triangle3f ta, vector3f na, triangle3f tb, vector3f nb,
sectline *sectarray);
/* clips triangles in mesh 'pmeshin' according to clipping
lines in 'pline'   adds new triangles to 'pmeshout' */
int ClipTriangles(mesh2 *pmeshin, mesh2 *pmeshout, sectline *pline, int
linenum, int flipnormals);

/* clips triangle of 'pmeshin' with index 'trindex' according to clipping
  lines in 'pline'
  adds new triangles to 'pmeshout' */
int ClipTriangle (mesh2 *pmeshin, mesh2 *pmeshout, int trindex, sectline *pline,
int linenum, int flipnormals);

  /* computes the triangles of a union, and adds it's triangles to 'pnewmesh' */
int Bool (mesh2 *pamesh, mesh2 *pbmesh, mesh2 *pnewmesh, int what);
  /* finds a line to continue 'ppreviusline->endpoint[1]'
   with and puts it into 'plineout'
  the line is either from the 'linenum' long 'pline' array, or a segment of a
  side of 'triangle' --- in the latter case 'plineout->flag' & SIDE = true
  if no line found, 'pline->endtype[0]' = none
  'plineout->endpoint[0]' = 'ppreviousline->endpoint[1]'

  if returned line is side "n" of 'ptriangle', then sets the NEIGHBOUR"n" flag
  of 'ptriangle'
  */
int selectnextline(sectline *pline, int linenum, mtriangle *ptriangle, sectline
*ppreviousline, sectline *plineout, matrix4x4f transformmatrix);
  /* adds those neighbours of triangle in 'pmeshin' indexed with 'trindex'
  which are needed, and not touched (these are indicated by their flags)
  recursively */
int AddNeededNeighbours(mtriangle *ptriangle, mesh2 *pmeshin, mesh2 *pmeshout);
  /* adds all untouched neighbour */
int AddAllNeighbours(mtriangle *ptriangle, mesh2 *pmeshin, mesh2 *pmeshout);

  /* adds vertex b to vertex a */
//int vertexadd3f(vertex3f a, vertex3f b);

#define vertexadd3f(a,b) ((a)[0]+=(b)[0], (a)[1]+=(b)[1], (a)[2]+=(b)[2])
#define vertexsub3f(a,b) ((a)[0]-=(b)[0], (a)[1]-=(b)[1], (a)[2]-=(b)[2])
#define vertexmul3f(a,b) ((a)[0]*=(b), (a)[1]*=(b), (a)[2]*=(b))
#define vertexdiv3f(a,b) ((a)[0]/=(b), (a)[1]/=(b), (a)[2]/=(b))
 /* returns true, if vectors a and b are (near) paralell */
int paralell3f (vector3f a, vector3f b);
 /* returns true, if vectors a and b are (near) the same */
int nearequal3f (vector3f a, vector3f b);
float distance3f (vertex3f a, vertex3f b);
 /* returns true, if the normals of 'a' and 'b' do not (both) point out of the
triangle described by 'a' and 'b' endpoints, assuming, that
'a.endpoint[1]' = 'b.endpoint[0]' */
int concave(sectline *a, sectline *b);
/* returns true, if 'a' and 'b' have their normals on the same side ,
assuming, that 'a.endpoint[1]' = 'b.endpoint[0]' */
int samesidenormal(sectline *a, sectline *b);
/* returns true, if the 'pline->endpoint[1]' -> 'vertex' vector points "above"
 pline */
int goodpoint(sectline *pline, vertex3f vertex);
  /* returns true, if there is a sectline in 'pline' without flag TOUCHED
  set, which intersects the line between pline[i].endpoint[0] and
  pline[j].endpoint[1] */
int islinebetween(sectline *pline, int linenum, vector3f normal, int i, int j);

/* return value is zero if a and b are paralell, and grows as their angle
 shrinks */
float howparalell (sectline *a, sectline *b);
#ifdef DEBUG
void printfvertex(vertex3f vertex);
void printftriangle(triangle3f triangle);
#else
#define printfvertex(a)
#define printftriangle(a)
#endif
#endif // first inclusion of thus header
