/* coordtest.c
  used to test coordgeom.c
  Prim Andras 2002
*/
#include <math.h>
#include <stdlib.h>
#include "coordgeom.h"
#include "pastuff.h"
#include "myio.h"
#include <GL/glut.h>


#define MESHBASEPOINTS 15
#define MESHTRIANGLES 4*MESHBASEPOINTS
#define NEARPLANE 2
#define MAXITER   10
#define BOTTOM 0
#define TOP    2
#define WIDTH 0.5

float lightpos[] = {-3.0, 2.0,  2.0, 1.0};
float lightdir[] = {1.0,  -1.0, -1.0 };

float lightdiff[] = {1.0, 1.0, 1.0, 1.0};
float lightspec[] = {1.0, 1.0, 1.0, 1.0};
float lightamb[] = {0.2, 0.2, 0.2, 1.0};

float matdiff[][4] = {{0.9, 0.2, 0.2, 1.0},
                      {0.4, 0.5, 0.7, 1.0}};
float matspec[][4] = {{0.9, 0.7, 0.4, 1.0},
                      {0.4, 0.5, 0.7, 1.0}};
float matamb[][4] = {{0.3, 0.07, 0.07, 1.0},
                     {0.7, 0.9, 1.0, 1.0}};

int xdown = -1, ydown = -1, colind = 0;
float transx = 0, transy = 0, transz = -1;
double fi = 0.0, fitmp = 0.0, teta = 0.0, tetatmp = 0.0;
double sigma = 0.0, sigmatmp = 0.0;
vector3f normal[3], sect[2];
triangle3f triangle1, triangle2;
mesh meshone = {0, NULL, NULL}, meshtwo = {0, NULL, NULL};
mesh2 mesh2one, mesh2two;
mesh2 *othermesh;
sectline *pline = NULL;
int linenum = 0;
float weight = 0;
vertex3f basepoints[MESHBASEPOINTS];
vertex3f toppoints[MESHBASEPOINTS];
char num[12] = {0,0,0,0,0,0,0,0,0,0};
float darkwire[3] = {0.5 , 0.5, 0.5};


void newmesh(void)
{
  int i, j;
  float width = WIDTH, deg = (float) 2 * M_PI / MESHBASEPOINTS;
  vertex3f vertextmp, lastvertex;
  vertex3f bmiddle = {0.0, BOTTOM, 0.0};
  vertex3f tmiddle = {0.0, TOP, 0.0};
  triangle3f  triangletmp;

debfuncbegin("newmesh\n");
  InitMesh(&mesh2one);
  InitMesh(&mesh2two);

  paPushMatrix();
  paLoadIdentity();
/*  paRotatef(-25, 1, 1, 1);
  paMatrixVertex3f(bmiddle, bmiddle);
  paMatrixVertex3f(tmiddle, tmiddle);
*/
  vertextmp[0] = 0.0;
  vertextmp[1] = BOTTOM;
  vertextmp[2] = width;
  paMatrixVertex3f(vertextmp, vertextmp);
  memcpy (basepoints[0], vertextmp, sizeof(vertex3f));
  vertextmp[0] = 0.0;
  vertextmp[1] = TOP;
  vertextmp[2] = width;
  paMatrixVertex3f(vertextmp, vertextmp);
  memcpy (toppoints[0], vertextmp, sizeof(vertex3f));

  for (i = 1, j = 0;j < MESHBASEPOINTS; i++, j++)
  {
    memcpy(lastvertex, vertextmp, sizeof(vertex3f));

    if( i == MESHBASEPOINTS)
    {
      i = 0;
//      memcpy (vertextmp, basepoints[0],  sizeof(vertex3f));
    }
    else
    {
      vertextmp[0] = sin(deg*i)*width;
      vertextmp[1] = BOTTOM;
      vertextmp[2] = cos(deg*i)*width;
      paMatrixVertex3f(vertextmp, vertextmp);
      memcpy (basepoints[i], vertextmp, sizeof(vertex3f));
      vertextmp[0] = sin(deg*i)*width;
      vertextmp[1] = TOP;
      vertextmp[2] = cos(deg*i)*width;
      paMatrixVertex3f(vertextmp, vertextmp);
      memcpy (toppoints[i], vertextmp, sizeof(vertex3f));
    }

    if(MESHBASEPOINTS != 4)
    {
      memcpy(triangletmp[0], basepoints[j], sizeof(vertex3f));
      memcpy(triangletmp[1], bmiddle, sizeof(vertex3f));
      memcpy(triangletmp[2], basepoints[i], sizeof(vertex3f));
      AddTriangleToMesh(&mesh2one, triangletmp);

//    GetMeshTriangle3f(&mesh2one, (i-1)*4, &triangletmp);

    AddTriangleToMesh(&mesh2two, triangletmp);


      memcpy(triangletmp[0], toppoints[j], sizeof(vertex3f));
      memcpy(triangletmp[1], toppoints[i], sizeof(vertex3f));
      memcpy(triangletmp[2], tmiddle, sizeof(vertex3f));

      AddTriangleToMesh(&mesh2one, triangletmp);
//  GetMeshTriangle3f(&mesh2one, (i-1)*4 + 1, &triangletmp);

    AddTriangleToMesh(&mesh2two, triangletmp);
    }

    memcpy(triangletmp[0], basepoints[j], sizeof(vertex3f));
    memcpy(triangletmp[1], toppoints[i], sizeof(vertex3f));
    memcpy(triangletmp[2], toppoints[j], sizeof(vertex3f));
    AddTriangleToMesh(&mesh2one, triangletmp);
//  GetMeshTriangle3f(&mesh2one, (i-1)*4 + 2, &triangletmp);
    AddTriangleToMesh(&mesh2two, triangletmp);

    memcpy(triangletmp[0], basepoints[j], sizeof(vertex3f));
    memcpy(triangletmp[1], basepoints[i], sizeof(vertex3f));
    memcpy(triangletmp[2], toppoints[i], sizeof(vertex3f));
    AddTriangleToMesh(&mesh2one, triangletmp);
//  GetMeshTriangle3f(&mesh2one, (i-1)*4 + 2, &triangletmp);

    AddTriangleToMesh(&mesh2two, triangletmp);

  }
  paPopMatrix();

  if (MESHBASEPOINTS == 4)
  {
    memcpy(triangletmp[0], basepoints[1], sizeof(vertex3f));
    memcpy(triangletmp[1], basepoints[2], sizeof(vertex3f));
    memcpy(triangletmp[2], basepoints[3], sizeof(vertex3f));
    AddTriangleToMesh(&mesh2one, triangletmp);

    memcpy(triangletmp[0], basepoints[1], sizeof(vertex3f));
    memcpy(triangletmp[1], basepoints[0], sizeof(vertex3f));
    memcpy(triangletmp[2], basepoints[3], sizeof(vertex3f));
    AddTriangleToMesh(&mesh2one, triangletmp);

    memcpy(triangletmp[0], toppoints[0], sizeof(vertex3f));
    memcpy(triangletmp[1], toppoints[1], sizeof(vertex3f));
    memcpy(triangletmp[2], toppoints[2], sizeof(vertex3f));
    AddTriangleToMesh(&mesh2one, triangletmp);

    memcpy(triangletmp[0], toppoints[0], sizeof(vertex3f));
    memcpy(triangletmp[1], toppoints[2], sizeof(vertex3f));
    memcpy(triangletmp[2], toppoints[3], sizeof(vertex3f));
    AddTriangleToMesh(&mesh2one, triangletmp);
  }

debfuncend("newmesh end\n");

}

void freetest (void)
{
  FreeStaticMesh(&mesh2one);
  FreeStaticMesh(&mesh2two);
  if(0 != meshone.trianglenum)
  {
    free(meshone.triangles);
    free(meshone.normals);
  }
  if(0 != meshtwo.trianglenum)
  {
    free(meshtwo.triangles);
    free(meshtwo.normals);
  }
  if (linenum != 0)
    free(pline);
}

void createothermesh(void)
{
  vertex3f vnull = {0.0, 0.0, 0.0};
  vertex3f base[MESHBASEPOINTS], basemid;
  vertex3f top[MESHBASEPOINTS], topmid;
  triangle3f triangletmp;
  int i, j;


  othermesh = NewMesh();
  paLoadIdentity();
  paTranslatef(transx, transy, transz);
  paTranslatef(0, BOTTOM, 0);
  paMatrixVertex3f(vnull,basemid);
  for (i = 0; i < MESHBASEPOINTS; i++)
    paMatrixVertex3f(basepoints[i],base[i]);
  paRotatef(fitmp, 0, 1, 0);
  paRotatef(tetatmp, 1, 0, 0);
  paRotatef(sigmatmp, 0, 0, 1);
  paTranslatef(0, TOP - BOTTOM, 0);
  paMatrixVertex3f(vnull,topmid);
  for (i = 0; i < MESHBASEPOINTS; i++)
    paMatrixVertex3f(basepoints[i],top[i]);

  for (i = 0, j = 1; i < MESHBASEPOINTS; i++, j = (j < MESHBASEPOINTS - 1 ?
    j+1 : 0))
  {
    memcpy(triangletmp[0], base[i], sizeof(vertex3f));
    memcpy(triangletmp[2], top[i], sizeof(vertex3f));
    memcpy(triangletmp[1], base[j], sizeof(vertex3f));
    AddTriangleToMesh(othermesh, triangletmp);

    memcpy(triangletmp[0], top[i], sizeof(vertex3f));
    memcpy(triangletmp[2], top[j], sizeof(vertex3f));
    memcpy(triangletmp[1], base[j], sizeof(vertex3f));
    AddTriangleToMesh(othermesh, triangletmp);

    memcpy(triangletmp[0], top[i], sizeof(vertex3f));
    memcpy(triangletmp[1], top[j], sizeof(vertex3f));
    memcpy(triangletmp[2], topmid, sizeof(vertex3f));
    AddTriangleToMesh(othermesh, triangletmp);

    memcpy(triangletmp[0], base[i], sizeof(vertex3f));
    memcpy(triangletmp[2], base[j], sizeof(vertex3f));
    memcpy(triangletmp[1], basemid, sizeof(vertex3f));
    AddTriangleToMesh(othermesh, triangletmp);

  }

  if (4 == MESHBASEPOINTS)
  {
    memcpy(triangletmp[0], top[0], sizeof(vertex3f));
    memcpy(triangletmp[2], top[2], sizeof(vertex3f));
    memcpy(triangletmp[1], top[1], sizeof(vertex3f));
    AddTriangleToMesh(othermesh, triangletmp);

    memcpy(triangletmp[0], top[0], sizeof(vertex3f));
    memcpy(triangletmp[2], top[3], sizeof(vertex3f));
    memcpy(triangletmp[1], top[2], sizeof(vertex3f));
    AddTriangleToMesh(othermesh, triangletmp);

    memcpy(triangletmp[0], base[0], sizeof(vertex3f));
    memcpy(triangletmp[2], base[1], sizeof(vertex3f));
    memcpy(triangletmp[1], base[2], sizeof(vertex3f));
    AddTriangleToMesh(othermesh, triangletmp);

    memcpy(triangletmp[0], base[0], sizeof(vertex3f));
    memcpy(triangletmp[2], base[2], sizeof(vertex3f));
    memcpy(triangletmp[1], base[3], sizeof(vertex3f));
    AddTriangleToMesh(othermesh, triangletmp);
  }
}


void main_display (void)
{
  int i, maxi;
  vertex3f vtmp;
  vector3f vectortmp;
  triangle3f triangletmp;
  mesh2 *pnewmesh, *subdivmesh;
debfuncbegin("main_display\n");

  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(0, -1, NEARPLANE-5);

  paOuterLoadIdentity(mesh2two.transformmatrix);
  paOuterTranslatef(mesh2two.transformmatrix, transx, transy, transz);
  paOuterRotatef(mesh2two.transformmatrix, fitmp, 0, 1, 0);
  paOuterRotatef(mesh2two.transformmatrix, tetatmp, 1, 0, 0);
  paOuterRotatef(mesh2two.transformmatrix, sigmatmp, 0, 0, 1);

//  DrawMeshGL(&mesh2one, DRAW_WIRE);

  createothermesh();

  colind = 1;
  glMaterialfv(GL_FRONT, GL_AMBIENT, matamb[colind]);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, matdiff[colind]);
  glMaterialfv(GL_FRONT, GL_SPECULAR, matspec[colind]);

//  DrawMeshGL(othermesh, DRAW_WIRE);
  memcpy (mesh2one.color, darkwire, sizeof(float)*3);
  memcpy (mesh2two.color, darkwire, sizeof(float)*3);
    DrawMeshGL(&mesh2one, DRAW_WIRE);
    DrawMeshGL(&mesh2two, DRAW_WIRE);
  colind = 0;
  glMaterialfv(GL_FRONT, GL_AMBIENT, matamb[colind]);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, matdiff[colind]);
  glMaterialfv(GL_FRONT, GL_SPECULAR, matspec[colind]);

  glDisable(GL_LIGHTING);
  glDisable(GL_DEPTH_TEST);
  pnewmesh = NewMesh();
  if (num[9])
    Bool(&mesh2two, &mesh2one, pnewmesh, INTERSECTION);
  else if(num[8])
    Bool(&mesh2two, &mesh2one, pnewmesh, AMINUSB);
  else
    Bool(&mesh2two, &mesh2one, pnewmesh, UNION);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);


/*  pnewmesh->color[0] = 1;
  pnewmesh->color[1] = 1;
  pnewmesh->color[2] = 1;
*/
//  glDisable(GL_LIGHTING);
//  glDisable(GL_DEPTH_TEST);
  colind = 1;
  glMaterialfv(GL_FRONT, GL_AMBIENT, matamb[colind]);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, matdiff[colind]);
  glMaterialfv(GL_FRONT, GL_SPECULAR, matspec[colind]);

  DrawMeshGL(pnewmesh, DRAW_SOLID);
/*  subdivmesh = NewMesh();
  SubDivide(&mesh2one, subdivmesh, 0.1, weight);
//  FreeMesh(pnewmesh);
  pnewmesh = NewMesh();
  SubDivide(subdivmesh, pnewmesh, 0.1, weight);
*/
/*  FreeMesh(subdivmesh);
  subdivmesh = NewMesh();
  SubDivide(pnewmesh, subdivmesh, 0.01, weight);
*/
  if (num[1])
    DrawMeshGL(&mesh2one, DRAW_WIRE);
  if (num[2])
    DrawMeshGL(&mesh2one, DRAW_SOLID);
  if (num[3])
    DrawMeshGL(subdivmesh, DRAW_WIRE);
  if (num[4])
    DrawMeshGL(subdivmesh, DRAW_SOLID);
  if (num[5])
    DrawMeshGL(pnewmesh, DRAW_WIRE);
  if (num[6])
    DrawMeshGL(pnewmesh, DRAW_SOLID);

//  FreeMesh(subdivmesh);
  colind = 0;
  glMaterialfv(GL_FRONT, GL_AMBIENT, matamb[colind]);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, matdiff[colind]);
  glMaterialfv(GL_FRONT, GL_SPECULAR, matspec[colind]);
//  glEnable(GL_DEPTH_TEST);
//  glEnable(GL_LIGHTING);

  FreeMesh(pnewmesh);
  FreeMesh(othermesh);

  glutSwapBuffers();
debfuncend("main_display end\n");

}

void main_keyboard (unsigned char key, int x, int y)
{
  switch(key)
  {
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      key -= '0';
      num[key] = 1-num[key];
      break;
    case 'n':
//      newtriangle();
//      newmesh();
      break;
    case 'd':
      transx += 0.2;
      break;
    case 'a':
      transx -= 0.2;
      break;
    case 'w':
      transy += 0.2;
      break;
    case 's':
      transy -= 0.2;
      break;
    case 'f':
      transz += 0.2;
      break;
    case 'r':
      transz -= 0.2;
      break;

    case 'l':
      fitmp += 0.5;
      break;
    case 'j':
      fitmp -= 0.5;
      break;
    case 'k':
      tetatmp += 0.5;
      break;
    case 'i':
      tetatmp -= 0.5;
      break;
    case 'u':
      sigmatmp += 0.5;
      break;
    case 'o':
      sigmatmp -= 0.5;
      break;
    case '+':
      weight += 0.003;
      break;
    case '-':
      weight -= 0.003;
      break;
  }

  glutPostRedisplay();
}



void main_motion(int x, int y)
{
  double a, c, g, h;
  if(xdown != -1)
  {

    fitmp = fi + (xdown - x);
    tetatmp = teta + (ydown - y);

    glutPostRedisplay();
  }
}

void main_mouse(int button, int state, int x, int y)
{
  if (button == GLUT_LEFT_BUTTON)
  {
    if (GLUT_DOWN == state)
    {
      xdown = x;
      ydown = y;
    }
    else
    {
      fi = fitmp;
      teta = tetatmp;
      xdown = -1;
      ydown = -1;
    }
  }
}

int main(int argc, char *args[])
{
  int window;
debfuncbegin("main\n");
  newmesh();

  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
  glutInitWindowSize(800,800);
  glutInitWindowPosition(50, 50);
  glutInit(&argc, args);

  window = glutCreateWindow("Union");
  glutDisplayFunc(main_display);
  glutMouseFunc(main_mouse);
  glutMotionFunc(main_motion);
  glutKeyboardFunc(main_keyboard);

  glEnable(GL_DEPTH_TEST);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
  glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, lightdir);

  glLightfv(GL_LIGHT0, GL_AMBIENT, lightamb);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightdiff);
  glLightfv(GL_LIGHT0, GL_SPECULAR, lightspec);
  glLighti(GL_LIGHT0, GL_SPOT_EXPONENT, 30);
  glLighti(GL_LIGHT0, GL_SPOT_CUTOFF, 180);
  glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
  glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0);
  glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.0);

  glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 0.0);
  glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 0.0);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lightamb);

  glMaterialf(GL_FRONT, GL_SHININESS, 3);
  glMaterialfv(GL_FRONT, GL_AMBIENT, matamb[colind]);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, matdiff[colind]);
  glMaterialfv(GL_FRONT, GL_SPECULAR, matspec[colind]);


  glMatrixMode(GL_PROJECTION);
  glFrustum(-1, 1,  -1,  1,  NEARPLANE,  NEARPLANE + 100);

  glutPostRedisplay();
  glutMainLoop();
  freetest();

debfuncend("main end\n");
  return 0;
} // main()
