/* main.c
Prim Andras 2002

The RenderMan(R) Interface Procedures and Protocol are:
 Copyright 1988,1999,2000 Pixar
 All Rights Reserved
*/

#include "lsystem.h"
#include "myio.h"
#include "turtle3d.h"

#include <GL/glut.h>

#define NEARPLANE 2
#define FARPLANE 100
#define GRU  2

#define MAXITER   20
//#define RIBFILENAME "fa.rib"
//#define IMAGEFILENAME "fa.tif"

#define MINLENGTH  0.02
#define W_DEF      -0.01

char ribfilename[] = "tree.rib";
char imagefilename[] = "tree.tif";
char texbg[] = "texworld.tif";
char texground[] = "texground.tif";
char texbranch[] = "texbranch.tif";
char texleaf[] = "texleaf.tif";

float lightpos[] = {-5.0, 2.0,  2.0, 1.0};
float lightdir[3]; // = {1.0,  -1.0, -1.0 };

float lightdiff[] = {1.0, 1.0, 1.0, 1.0};
float lightspec[] = {1.0, 1.0, 1.0, 1.0};
float lightamb[] = {0.2, 0.2, 0.2, 1.0};

float matdiff[][4] = {{0.9, 0.2, 0.2, 1.0},
                      {0.2, 0.8, 0.6, 1.0}};
float matspec[][4] = {{0.9, 0.7, 0.4, 1.0},
                      {0.2, 0.8, 0.6, 1.0}};
float matamb[][4] = {{0.3, 0.07, 0.07, 1.0},
                     {0.1, 0.35, 0.3, 1.0}};

char num[10] = {1,0,0,0,0,0,0,0,0,0};

int xdown = -1, ydown = -1, colind = 0, iteration = 2;
float transx = 0, transy = -2, transz = NEARPLANE - 20;
double fi = 0.0, fitmp = 0.0, teta = 0.0, tetatmp = 0.0;
plant pl;
lsystem *ls;
#ifndef OLDDRAWTREE
/*vertex3f *vertexes;
vector3f *normals;
int trianglenum = 0;*/
mesh2 *ptreemesh, *pleafmesh, *subdivmesh = NULL, *subdivmesh2 = NULL;
#endif

void WriteRibFile ()
{
  FILE *fil;
  int i, j;
  fil = fopen(ribfilename, "wt");
  if (NULL == fil)
    return;

  fputs("# tree generated by a tree modeller of Prim Andras\n", fil);
  fprintf(fil, "Display \"%s\" \"file\" \"rgb\"\n", imagefilename);
  fputs("Format 400 400 1\n", fil);
  fputs("Projection \"perspective\"\n", fil);
  fputs("PixelSamples 1 1\n", fil);
  fputs("Attribute \"light\" \"shadows\" [\"on\"] \n", fil);

  fprintf(fil, "LightSource \"distantlight\" 1 \"from\" [%f %f %f]"
    "\"to\" [%f %f %f] \n", lightpos[0], lightpos[1], -lightpos[2],
    transx, transy, -transz);
  fputs("\nWorldBegin\n", fil);
  fprintf(fil, "Translate %f %f %f\n", transx, transy, -transz);

// Background
  fputs("AttributeBegin\n", fil);
  fputs("# background\n", fil);
  fputs("  Color [0.6 0.8 0.8]\n", fil);
  fprintf(fil, "  Surface \"paintedplastic\" \"Kd\" [2] \"Ks\" [0.5] "
   "\"texturename\" \"%s\"\n", texbg);
  fputs("\n  Polygon \"P\" [-150 -150 100  -150 150 100  150 150 100 "
    " 150 -150 100]", fil);
  fputs("\n    \"st\" [0 0  0 1  1 1  1 0]\n", fil);
  fputs("AttributeEnd\n\n", fil);

/* this flips z axis (since OpenGL's z axis points outward,
 RenerMan's points inward)
 and also enlarge to compensate difference */
  fprintf(fil, "Scale 1 1 -1\n");
  fprintf(fil, "Rotate %f 0 1 0\n", fitmp);
  fprintf(fil, "Rotate %f 1 0 0\n", tetatmp);

// ground
  fputs("AttributeBegin\n", fil);
  fputs("# ground\n", fil);
  fputs("  Attribute \"render\" \"visibility\" [7]\n", fil);

  fputs("  Color [0.2 0.5 0.3]\n", fil);
  fprintf(fil, "  Surface \"paintedplastic\" \"Kd\" [4] \"texturename\""
     " \"%s\"\n", texground);
  fputs("\n  Polygon \"P\" [-30 0 -30  30 0 -30  30 0 30  -30 0 30]", fil);
  fputs("\n    \"st\" [0 0  0 1  1 1  1 0]\n", fil);
  fputs("AttributeEnd\n\n", fil);

// branches
  fputs ("\nAttributeBegin\n",fil);
  fputs("# branches\n", fil);
  fputs("  Attribute \"render\" \"visibility\" [7]\n", fil);
  fputs("  Attribute \"render\" \"casts_shadows\" [\"Os\"]\n", fil);
  fputs ("  Color [1 0.9 0.7]\n", fil);
  fputs ("  Opacity [1 1 1]\n", fil);
  fprintf(fil, "  Surface \"paintedplastic\" \"Kd\" [2] \"Ks\" [0]"
     " \"texturename\""      " \"%s\"\n", texbranch);

  if(num[1])
  {
    MeshToRib(subdivmesh, fil, "  ");
    printf("Sub-division mesh iteration 1 written to %s\n", ribfilename);
  }
  else if(num[2])
  {
    MeshToRib(subdivmesh2, fil, "  ");
    printf("Sub-division mesh iteration 2 written to %s\n", ribfilename);
  }
  else
  {
    MeshToRib(ptreemesh, fil, "  ");
    printf("Normal mesh written to %s\n", ribfilename);
  }
  fputs ("AttributeEnd\n",fil);

// leaves
  fputs ("AttributeBegin\n",fil);
  fputs("# leaves\n", fil);
  fputs("  Attribute \"render\" \"visibility\" [3]\n", fil);
  fputs("  Attribute \"render\" \"casts_shadows\" [\"Os\"]\n", fil);
  fputs("  Sides 2\n", fil);

  fputs ("  Color [0.2 1 0.5]\n", fil);
  fputs ("  Opacity [0.8 0.8 0.8]\n", fil);
  fprintf(fil, "  Surface \"paintedplastic\" \"Kd\" [2] \"texturename\""
     " \"%s\"\n", texleaf);
  MeshToRib(pleafmesh, fil, "  ");
  fputs ("AttributeEnd\n",fil);
  fputs("WorldEnd\n", fil);
  fclose(fil);
}

void main_display (void)
{
  int i, j;
  vertex3f vtmp;

  glClearColor(0.2, 0.3, 0.5, 0.0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(transx, transy, transz);
  glRotated(fitmp, 0, 1, 0);
  glRotated(tetatmp, 1, 0, 0);

#ifndef OLDDRAWTREE
  colind = 0;
  glMaterialfv(GL_FRONT, GL_AMBIENT, matamb[colind]);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, matdiff[colind]);
  glMaterialfv(GL_FRONT, GL_SPECULAR, matspec[colind]);

  if(num[1])
  {
    DrawMeshGL(subdivmesh, DRAW_SOLID);
  }
  else if(num[3])
  {
    DrawMeshGL(subdivmesh, DRAW_DEBUG_WIRE);
  }
  else if(num[2])
  {
    DrawMeshGL(subdivmesh2, DRAW_SOLID);
  }
  else if(num[4])
    DrawMeshGL(ptreemesh, DRAW_WIRE);
  else
    DrawMeshGL(ptreemesh, DRAW_SOLID);

  colind = 1;
  glMaterialfv(GL_FRONT, GL_AMBIENT, matamb[colind]);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, matdiff[colind]);
  glMaterialfv(GL_FRONT, GL_SPECULAR, matspec[colind]);

  if (num[4] || num[3])
    DrawMeshGL(pleafmesh, DRAW_WIRE);
  else
    DrawMeshGL(pleafmesh, DRAW_SOLID);
#else
  OldDrawTree(pl.sentence);
//  DrawTree(pl.sentence);
#endif // OLDDRAWTREE
  glutSwapBuffers();
}


void main_motion(int x, int y)
{
  double a, c, g, h;
  if(xdown != -1)
  {

    fitmp = fi + (xdown - x);
    tetatmp = teta + (ydown - y);

    glutPostRedisplay();
  }
}

void main_mouse(int button, int state, int x, int y)
{
  if (button == GLUT_LEFT_BUTTON)
  {
    if (GLUT_DOWN == state)
    {
      xdown = x;
      ydown = y;
    }
    else
    {
      fi = fitmp;
      teta = tetatmp;
      xdown = -1;
      ydown = -1;
    }
  }
}


void main_keyboard (unsigned char key, int x, int y)
{
  int modified = 0, i;
  float light = 0.5;
  glPushMatrix();
  glLoadIdentity();

  switch(key)
  {
    case '+':
    case '=':
      if (iteration < MAXITER)
      {
        free(pl.sentence);
        LsInitPlant(&pl, ls);
        iteration++;
        modified = 1;
      }
      break;
    case '-':
    case '_':
      if (iteration > 0)
      {
        free(pl.sentence);
        LsInitPlant(&pl, ls);
        iteration--;
        modified = 1;
      }
      break;
    case 'd':
      transx += 1;
      break;
    case 'a':
      transx -= 1;
      break;
    case 'w':
      transy += 1;
      break;
    case 's':
      transy -= 1;
      break;
    case 'f':
      transz += 1;
      break;
    case 'r':
      transz -= 1;
      break;
    case '?':
      WriteRibFile();
      break;
    case '[':
      light = -0.5;
    case ']':
      lightpos[0] += light;
      glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
  lightdir[0] = -lightpos[0];
  glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, lightdir);
      break;
    case '/':
      light = -0.5;
    case '\'':
      lightpos[1] += light;
      glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
  lightdir[1] = -lightpos[1];
  glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, lightdir);
      break;
    case ';':
      light = -0.5;
    case '.':
      lightpos[2] += light;
      glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
  lightdir[2] = -lightpos[2];
  glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, lightdir);
      break;
    case '1':
    case '2':
    case '3':
      if (NULL == subdivmesh)
      {
        subdivmesh = NewMesh();
        SubDivide(ptreemesh, subdivmesh, MINLENGTH, W_DEF);
      }
      if ((key == '2') && (NULL == subdivmesh2))
      {
        subdivmesh2 = NewMesh();
        SubDivide(subdivmesh, subdivmesh2, MINLENGTH, W_DEF);
      }
    case '0':
    case '4':
/*    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
*/
      key -= '0';
      memset(num, 0, 10);
      num[key] = 1;
      break;
  }

  glPopMatrix();
  if(modified)
  {
    for (i = 0; i < iteration; i++)
      LsDerive(&pl);
#ifndef OLDDRAWTREE
/*    free(vertexes);
    free(normals);
    trianglenum = CreateTreeMesh(&pl, &vertexes, &normals);
    glVertexPointer(3, GL_FLOAT, trianglenum*3, vertexes);
    glNormalPointer(GL_FLOAT, trianglenum*3, normals);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY); */
//  LsPrintSentence(&pl);
  FreeMesh(ptreemesh);
  FreeMesh(pleafmesh);
  ptreemesh = NewMesh();
  pleafmesh = NewMesh();
  CreateLeafyTreeMesh(&pl, ptreemesh, pleafmesh);

  if (NULL != subdivmesh)
  {
    FreeMesh(subdivmesh);
    subdivmesh = NULL;
  }
  if (NULL != subdivmesh2)
  {
    FreeMesh(subdivmesh2);
    subdivmesh2 = NULL;
  }
  if (num[1] || num[2])
  {
    subdivmesh = NewMesh();
    SubDivide(ptreemesh, subdivmesh, MINLENGTH, W_DEF);
    if (num[2])
    {
      subdivmesh2 = NewMesh();
      SubDivide(subdivmesh, subdivmesh2, MINLENGTH, W_DEF);
    }
  }


#endif
  }
  glutPostRedisplay();
}


int main(int argc, char *args[])
{
  FILE *fil;
  int i;
  int window;

  if (argc < 2)
  {
    fprintf(stderr, "usage %s <lsfile>\n", args[0]);
    return 1;
  }

  if (NULL == (fil = fopen(args[1], "rt")))
  {
    fprintf(stderr, "%s: cannot open %s \n", args[0], args[1]);
    return 1;
  }

  if (NULL == (ls = LsReadFile (fil)))
  {
    fprintf(stderr, "%s: syntax error in %s \n", args[0], args[1]);
    return 1;
  }
  fclose(fil);
  LsInitPlant (&pl, ls);

  for (i = 0; i < iteration; i ++)
    LsDerive(&pl);
#ifndef OLDDRAWTREE
//  trianglenum = CreateTreeMesh(&pl, &vertexes, &normals);
  ptreemesh = NewMesh();
  pleafmesh = NewMesh();
  CreateLeafyTreeMesh(&pl, ptreemesh, pleafmesh);
  subdivmesh = NewMesh();
  subdivmesh2 = NewMesh();
//  SubDivide(ptreemesh, subdivmesh, MINLENGTH, W_DEF);
//  SubDivide(subdivmesh, subdivmesh2, MINLENGTH, W_DEF);
#endif

  glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
  glutInitWindowSize(800,800);
  glutInitWindowPosition(50, 50);
  glutInit(&argc, args);

  window = glutCreateWindow("The ultimate treedisplayer");
  glutDisplayFunc(main_display);
  glutMouseFunc(main_mouse);
  glutMotionFunc(main_motion);
  glutKeyboardFunc(main_keyboard);

  glEnable(GL_DEPTH_TEST);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT0, GL_POSITION, lightpos);

  lightdir[0] = -lightpos[0];
  lightdir[1] = -lightpos[1];
  lightdir[2] = -lightpos[2];
  glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, lightdir);

  glLightfv(GL_LIGHT0, GL_AMBIENT, lightamb);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightdiff);
  glLightfv(GL_LIGHT0, GL_SPECULAR, lightspec);
  glLighti(GL_LIGHT0, GL_SPOT_EXPONENT, 30);
  glLighti(GL_LIGHT0, GL_SPOT_CUTOFF, 180);
  glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
  glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.0);
  glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.0);

  glLightModelf(GL_LIGHT_MODEL_LOCAL_VIEWER, 0.0);
  glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 0.0);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lightamb);

  glMaterialf(GL_FRONT, GL_SHININESS, 3);
  glMaterialfv(GL_FRONT, GL_AMBIENT, matamb[colind]);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, matdiff[colind]);
  glMaterialfv(GL_FRONT, GL_SPECULAR, matspec[colind]);

  glMatrixMode(GL_PROJECTION);
  glFrustum(-GRU, GRU,  -GRU,  GRU,  NEARPLANE,  FARPLANE);

  glutPostRedisplay();
  glutMainLoop();

#ifndef OLDDRAWTREE
/*  free(vertexes);
  free(normals); */
  FreeMesh(ptreemesh);
  FreeMesh(pleafmesh);
  if (NULL != subdivmesh)
    FreeMesh(subdivmesh);
  if (NULL != subdivmesh2)
    FreeMesh(subdivmesh2);

#endif
  free(pl.sentence);
  freeLS(ls);
  return 0;
} // main()
