/*
mesh2.c
create, modify, query and deleting mesh2
Prim Andras 2002
*/

#include <stdlib.h>
#include "myio.h"
#include "mesh2.h"
#include "coordgeom.h"

#ifdef COORDTEST
#include <GL/glut.h>
#endif

int InitMesh(mesh2 *pmesh)
{
debfuncbegin("InitMesh\n");
  pmesh->trianglestore = newstore (sizeof(mtriangle));
  pmesh->vertexstore = newstore (sizeof(mvertex));
  pmesh->linestore = newstore (sizeof(mline));
  pmesh->vertexindexroot = NULL;
  pmesh->color[0] = 0.8; pmesh->color[1] = 0.7; pmesh->color[2] = 0.3;
  paOuterLoadIdentity(pmesh->transformmatrix);
debfuncend("InitMesh end\n");
  return 0;
}

 /* It inits the new mesh as well */
mesh2* NewMesh()
{
debfuncbegin("NewMesh\n");
  mesh2 *ret;
  ret = malloc(sizeof(mesh2));
  ret->trianglestore = newstore (sizeof(mtriangle));
  ret->vertexstore = newstore (sizeof(mvertex));
  ret->linestore = newstore (sizeof(mline));
  ret->vertexindexroot = NULL;
  ret->color[0] = 0.8; ret->color[1] = 0.7; ret->color[2] = 0.3;
  paOuterLoadIdentity(ret->transformmatrix);
debfuncend("NewMesh end\n");

  return ret;
}

  /* does not free the mesh2 structure, however, initialises it */
int FreeStaticMesh(mesh2 *pmesh)
{
debfuncbegin("FreeStaticMesh\n");
  freestore(pmesh->trianglestore);
  freestore(pmesh->vertexstore);
  freestore(pmesh->linestore);

  if (NULL != pmesh->vertexindexroot)
    FreeVertexIndexNode(pmesh->vertexindexroot);
  pmesh->vertexindexroot = NULL;
debfuncend("FreeStaticMesh end\n");
  return 0;
}


int FreeMesh(mesh2 *pmesh)
{
debfuncbegin("FreeMesh\n");
  freestore(pmesh->trianglestore);
  freestore(pmesh->vertexstore);
  freestore(pmesh->linestore);

  if (NULL != pmesh->vertexindexroot)
    FreeVertexIndexNode(pmesh->vertexindexroot);
  free(pmesh);
debfuncend("FreeMesh end\n");
  return 0;
}

  /* creates new mesh and copies the _pointers_ of pmesh in it */
mesh2 *CopyMeshUnsafe(mesh2 *pmesh)
{
  mesh2 *ret;
  ret = malloc(sizeof(mesh2));
  memcpy(ret, pmesh, sizeof(mesh2));
  return ret;
}

  /* exchange the contets of the two meshes */
int ExchangeMesh(mesh2 *pamesh, mesh2 *pbmesh)
{
  matrix4x4f matrixtmp;
  store *pstoretmp;
  vertexindexnode *pvertexindexnodetmp;
  vector3f vectortmp;

  pstoretmp = pamesh->trianglestore;
  pamesh->trianglestore = pbmesh->trianglestore;
  pbmesh->trianglestore = pstoretmp;

  pstoretmp = pamesh->vertexstore;
  pamesh->vertexstore = pbmesh->vertexstore;
  pbmesh->vertexstore = pstoretmp;

  pstoretmp = pamesh->linestore;
  pamesh->linestore = pbmesh->linestore;
  pbmesh->linestore = pstoretmp;

  pvertexindexnodetmp = pamesh->vertexindexroot;
  pamesh->vertexindexroot = pbmesh->vertexindexroot;
  pbmesh->vertexindexroot = pvertexindexnodetmp;

  memcpy(matrixtmp, pamesh->transformmatrix, sizeof(matrix4x4f));
  memcpy(pamesh->transformmatrix, pbmesh->transformmatrix, sizeof(matrix4x4f));
  memcpy(pbmesh->transformmatrix, matrixtmp, sizeof(matrix4x4f));

  memcpy(vectortmp, pamesh->color, sizeof(vector3f));
  memcpy(pamesh->color, pbmesh->color, sizeof(vector3f));
  memcpy(pbmesh->color, vectortmp, sizeof(vector3f));
}

  /* frees node and it's offsprings recursively */
int FreeVertexIndexNode(vertexindexnode *pnode)
{
  int i;
debfuncbegin("FreeVertexIndexNode\n");
  for(i = 0; i < 8; i++)
    if (NULL != pnode->offsprings[i])
      FreeVertexIndexNode(pnode->offsprings[i]);

  free(pnode);
debfuncend("FreeVertexIndexNode end\n");
  return 0;
}

int MeshToOldFormat(mesh2 *pmeshin, mesh *pmeshout)
{
debfuncbegin("MeshToOldFormat\n");
  int trianglenum, i;
  trianglenum = pmeshout->trianglenum =
    getstoreitemnum(*pmeshin->trianglestore);

  pmeshout->triangles = (triangle3f *) malloc(trianglenum*sizeof(triangle3f));
  pmeshout->normals = (vector3f *) malloc(trianglenum*sizeof(vector3f));

  for (i = 0; i < trianglenum; i++)
  {
    GetMeshTriangleAndNormal3f(pmeshin, i, pmeshout->triangles[i],
      pmeshout->normals[i]);
  }
debfuncend("MeshToOldFormat end\n");

  return 0;
}


  /* puts the triangle with index 'trindex' of pmesh into 'ptriangle' in
  triangle3f form */
int GetMeshTriangle3f (mesh2 *pmesh, int trindex, triangle3f ptriangle)
{
  int i;
//  vertex3f vertextmp;
  mtriangle *pmtriangle;
debfuncbegin("GetMeshTriangle3f\n");
  pmtriangle = getstoreitemaddress(pmesh->trianglestore, trindex);

debinfo2("pmtriangle %p\n", pmtriangle);
  if(NULL == pmtriangle)
  {
    memset(ptriangle, 0, sizeof(triangle3f));
debfuncend("GetMeshTriangle3f abnormal end\n");
    return 1;
  }
  for (i = 0; i < 3; i++)
  {
    paOuterMatrixVertex3f(pmesh->transformmatrix,
      pmtriangle->angles[i]->coord, ptriangle[i]);
//    memcpy(ptriangle[i], vertextmp, sizeof(vertex3f));
//  memcpy(&(*ptriangle)[i], pmtriangle->angles[i]->coord, sizeof(vertex3f));
  }

debfuncend("GetMeshTriangle3f end\n");
  return 0;
}


  /* puts the normal with index 'nindex' of pmesh into 'pnormal' in
  vector3f form */
int GetMeshNormal3f (mesh2 *pmesh, int nindex, vector3f pnormal)
{
  mtriangle *pmtriangle;
//  vector3f vectortmp;

debfuncbegin("GetMeshNormal3f\n");
  pmtriangle = getstoreitemaddress(pmesh->trianglestore, nindex);
  if(NULL == pmtriangle)
  {
    memset(pnormal, 0, sizeof(vector3f));
debfuncend("GetMeshNormal3f abnormal end\n");
    return 1;
  }
  paOuterMatrixVector3f(pmesh->transformmatrix,
    pmtriangle->normal, pnormal);
//  memcpy(pnormal, &vectortmp, sizeof(vector3f));
debfuncend("GetMeshNormal3f end\n");
  return 0;
}

int GetMeshTriangleAndNormal3f (mesh2 *pmesh, int trindex, triangle3f ptriangle,
  vector3f pnormal)
{
  int i;
//  vertex3f vertextmp;
  mtriangle *pmtriangle;
//  vector3f vectortmp;

//debfuncbegin("GetMeshTriangleAndNormal3f\n");
  pmtriangle = getstoreitemaddress(pmesh->trianglestore, trindex);

  if(NULL == pmtriangle)
  {
    memset(ptriangle, 0, sizeof(triangle3f));
    memset(pnormal, 0, sizeof(vector3f));
//debfuncend("GetMeshTriangleAndNormal3f abnormal end\n");
    return 1;
  }
  for (i = 0; i < 3; i++)
  {
    paOuterMatrixVertex3f(pmesh->transformmatrix,
      pmtriangle->angles[i]->coord, ptriangle[i]);
  }
  paOuterMatrixVector3f(pmesh->transformmatrix,
    pmtriangle->normal, pnormal);

//debfuncend("GetMeshTriangleAndNormal3f end\n");
  return 0;
}


int GetMeshLine3f(mesh2 *pmesh, int index, line3f pline)
{
  mline *pmline;

  pmline = getstoreitemaddress(pmesh->linestore, index);

  if(NULL == pmline)
  {
    memset(pline, 0, sizeof(line3f));
    return 1;
  }

  paOuterMatrixVertex3f(pmesh->transformmatrix,
    pmline->pvertex[0]->coord, pline[0]);
  paOuterMatrixVertex3f(pmesh->transformmatrix,
    pmline->pvertex[1]->coord, pline[1]);

//debfuncend("GetMeshTriangleAndNormal3f end\n");
  return 0;

}


int GetMeshTrianglesAndNormals3f(mesh2 *pmesh, vertex3f **vertexes, vector3f
**normals)
{
  int trianglenum, i, j, k;
  mtriangle *pmtriangle;
/*  triangle3f triangletmp;*/
  vector3f normaltmp;
//  vertex3f *apvertex[3];

  trianglenum = GetMeshTriangleNum(pmesh);
  *vertexes = (vertex3f*) malloc(trianglenum*3*sizeof(vertex3f));
  *normals = (vector3f*) malloc(trianglenum*3*sizeof(vector3f));

  for (i = 0, j = 0; i < trianglenum; i++)
  {
    pmtriangle = getstoreitemaddress(pmesh->trianglestore, i);

    paOuterMatrixVector3f(pmesh->transformmatrix,
      pmtriangle->normal, normaltmp);
/*    apvertex[0] = &pmtriangle->angles[0]->coord;
    apvertex[1] = &pmtriangle->angles[1]->coord;
    apvertex[2] = &pmtriangle->angles[2]->coord;
*/
    for (k = 0; k < 3; k++, j++)
    {
      paOuterMatrixVertex3f(pmesh->transformmatrix,
        pmtriangle->angles[k]->coord, (*vertexes)[j]);
      memcpy((*normals)[j], normaltmp, sizeof(vector3f));
    }
  }
  return trianglenum;
}


  /* adds the triangle to the mesh (allocates space for it, and fills
  values) */
int AddTriangleToMesh(mesh2 *pmesh, triangle3f ptriangle)
{
  int i, j, possible01 = 1, possible02 = 1, possible12 = 1;
  mtriangle *newtriangle, dummytriangle;
  mvertex *pvertex;
debfuncbegin("AddTriangleToMesh\n");
printftriangle(ptriangle);

  if ( nearequal3f(ptriangle[0], ptriangle[1]) ||
    nearequal3f(ptriangle[0], ptriangle[2]) ||
    nearequal3f(ptriangle[1], ptriangle[2]) )
  {
/*    printf ("abnormal triangle\n");
printf("[%f %f %f  %f %f %f  %f %f %f]\n",
  ptriangle[0][0], ptriangle[0][1], ptriangle[0][2],
  ptriangle[1][0], ptriangle[1][1], ptriangle[1][2],
  ptriangle[2][0], ptriangle[2][1], ptriangle[2][2]);
*/
    return 1;
  }
  newtriangle = addstoreitem (pmesh->trianglestore, &dummytriangle);
debinfo2("newtriangle: %p\n", newtriangle);
  memset(newtriangle->neighbours, 0, 3*sizeof(mtriangle*));
  newtriangle->flag = 0;
  newtriangle->trindex = getstoreitemnum(*pmesh->trianglestore) - 1;
    // sets pointers to angles
  for (i = 0; i < 3; i++)
  {
//printf("in %d (%f %f %f)\n",i,
//ptriangle[1][0],ptriangle[1][1],ptriangle[1][2]);
    pvertex = VertexPointer(pmesh, ptriangle[i]);
    newtriangle->angles[i] = pvertex;

    if (NULL == pvertex->pfirstline)
    {
        /* if  this is the first triangle in this vertex,
        we cannot hope a neighbouring triangle with it */
      switch(i)
      {
        case 0:
          possible01 = 0;
          possible02 = 0;
          break;
        case 1:
          possible01 = 0;
          possible12 = 0;
          break;
        case 2:
          possible02 = 0;
          possible12 = 0;
          break;
      }
    }
    else
    {
        /* copy back, since the vertex found is only "near" the same and
         we need the normal, and I don't want errors to accumulate */
      memcpy(ptriangle[i], pvertex->coord, sizeof(vertex3f));
    }
  }

//printf("2 (%f %f %f)\n",ptriangle[1][0],ptriangle[1][1],ptriangle[1][2]);
  GetNormal3f (ptriangle, newtriangle->normal);

  //  find neighbour triangles
  if (possible01)
  {
    FindNeighbourTriangleAndAdd (pmesh, newtriangle,0, 1);
  }
  else
  {
    NewMeshLine(pmesh, newtriangle, 0, 1);
    newtriangle->neighbours[0] = NULL;
  }
  if (possible02)
  {
    FindNeighbourTriangleAndAdd (pmesh, newtriangle,0, 2);
  }
  else
  {
    NewMeshLine(pmesh, newtriangle, 0, 2);
    newtriangle->neighbours[1] = NULL;
  }
  if (possible12)
  {
    FindNeighbourTriangleAndAdd (pmesh, newtriangle,1, 2);
  }
  else
  {
    NewMeshLine(pmesh, newtriangle, 1, 2);
    newtriangle->neighbours[2] = NULL;
  }
debfuncend("AddTriangleToMesh end\n");

  return 0;
}

int NewMeshLine(mesh2 *pmesh, mtriangle *ptriangle, int a, int b)
{
  char *index;
  mline linetmp, *plinetmp, *plastline, **pplinetmp;
  int itmp;
debfuncbegin("NewMeshLine\n");

  if(a > b)
  {itmp = a; a = b; b = itmp;}

  itmp = (a == 0 ? (b == 1 ? 0 : 1) : 2 );

  linetmp.pvertex[0] = ptriangle->angles[a];
  linetmp.pvertex[1] = ptriangle->angles[b];
  linetmp.ptriangle[0] = ptriangle;
  linetmp.trside[0] = itmp;
  linetmp.ptriangle[1] = NULL;
  linetmp.trside[1] = -1;
  linetmp.pnextline[0] = NULL;
  linetmp.pnextline[1] = NULL;

  plinetmp = addstoreitem(pmesh->linestore, &linetmp);

  pplinetmp = &(ptriangle->angles[a]->pfirstline);
  index = &(ptriangle->angles[a]->vindex);

  while(NULL != *pplinetmp)
  {
    plastline = (*pplinetmp);
    pplinetmp = &(plastline->pnextline[(*index)]);
    index = &(plastline->vindex[(*index)]);
  }

  (*pplinetmp) = plinetmp;
  (*index) = 0;


  pplinetmp = &(ptriangle->angles[b]->pfirstline);
  index = &(ptriangle->angles[b]->vindex);

  while(NULL != *pplinetmp)
  {
    plastline = (*pplinetmp);
    pplinetmp = &(plastline->pnextline[(*index)]);
    index = &(plastline->vindex[(*index)]);
  }

  (*pplinetmp) = plinetmp;
  (*index) = 1;

  ptriangle->pline[itmp] = plinetmp;


debfuncend("NewMeshLine end\n");

  return 0;
}


int AddPolygonToMesh(mesh2 *pmesh, sectline *pline, int
linenum, vector3f normal)
{
  int lines = linenum;
  int i = 0, j = 1, safety, besti, bestj;
  float at, bestat;
  triangle3f triangletmp;
  vector3f normaltmp;

debfuncbegin("AddPolygonToMesh\n");
//printf("linenum: %d\n", linenum);
  besti = i;
  bestj = j;
  while (lines > 2)
  {
    bestat = 0;
    safety = 0;
    do
    {
        // two untouched neighbouring sectline
       do{
         i++;
         i = i < linenum ? i : 0;
       }while(pline[i].flag & TOUCHED);
       j = i;
       do {
         j++;
         j = j < linenum ? j : 0;
       }while (pline[j].flag & TOUCHED);

       if (! ((concave(&pline[i], &pline[j])) ||
          islinebetween(pline, linenum, normal, i, j)))
       {
          at = howparalell(&pline[i], &pline[j]);
          if (at > bestat)
          {
            bestat = at;
            besti = i;
            bestj = j;
          }
       }

       safety++;
    }while((safety <= lines) );

/*    if (safety > linenum)
    {
debfuncend("AddPolygonToMesh abnormal end\n");
      return 1;
    }
*/
    i = besti;
    j = bestj;

    memcpy(triangletmp[0], pline[i].endpoint[0], sizeof(vertex3f));
    memcpy(triangletmp[1], pline[i].endpoint[1], sizeof(vertex3f));
    memcpy(triangletmp[2], pline[j].endpoint[1], sizeof(vertex3f));

    GetNormal3f(triangletmp, normaltmp);


      /* this hopefully means, that the order of the triangle's angles
       generates a normal directing to the opposite of the desired direction
       changing the angle order cures this */
    if (!nearequal3f(normal, normaltmp))
    {
      memcpy(triangletmp[1], pline[j].endpoint[1], sizeof(vertex3f));
      memcpy(triangletmp[2], pline[i].endpoint[1], sizeof(vertex3f));

      AddTriangleToMesh(pmesh, triangletmp);

      memcpy(triangletmp[2], pline[j].endpoint[1], sizeof(vertex3f));
    }
    else
    {
      AddTriangleToMesh(pmesh, triangletmp);
      memcpy(triangletmp[1], pline[j].endpoint[1], sizeof(vertex3f));
    }

      /* count the normal of the new sectline, created by joining
      pline[i].endpoint[0] and pline[j].endpoint[1] a bit below */
    vertexadd3f(triangletmp[1], normaltmp);
    GetNormal3f(triangletmp, pline[j].normal);
//    getnormalout(&pline[i], &pline[j], normaltmp);
//    memcpy(pline[j].normal, normaltmp, sizeof(vector3f));

    memcpy(pline[j].endpoint[0], pline[i].endpoint[0], sizeof(vertex3f));
    pline[i].flag |= TOUCHED;
    lines --;
  }

/*
glBegin(GL_LINES);
for( i = 0; i < linenum; i++)
{
  DrawLineAndNormalGL(&pline[i]);
}
glEnd();
*/
debfuncend("AddPolygonToMesh end\n");
  return 0;
}

  /* if mesh already contains a vertex similar to pvertex, returns it's address
   else inserts pvertex into the mesh, and returns the pointer
   it uses the space of pvertex in neither case  */
mvertex *VertexPointer(mesh2 *pmesh, vertex3f pvertex)
{
  vertexindexnode *nearest = NULL, *next;
  int index;
  vertex3f *coord;
  mvertex dummyvertex, *newvertex;
debfuncbegin("VertexPointer\n");
  if(getstoreitemnum(*pmesh->vertexstore) > 0)
  {
    if (NULL != pmesh->vertexindexroot)
    {
      next = pmesh->vertexindexroot;
      do
      {
        nearest = next;
        coord = &(nearest->item->coord);
        if (nearequal3f(pvertex, *coord))
        {
debfuncend("VertexPointer end 0\n");
          return nearest->item;
        }
        index = (pvertex[0] > (*coord)[0] ? 4 : 0) +
           (pvertex[1] > (*coord)[1] ? 2 : 0) +
           (pvertex[2] > (*coord)[2] ? 1 : 0);
        next = nearest->offsprings[index];
      }while(NULL != next);
    }
    else
    {
      nearest = NULL;
      next = pmesh->vertexindexroot;
    }
  }

  /* we reach this point, if no similar vertex found
  nearest->offsprings[index] is the place of the new vertexindexnode */
  newvertex = (mvertex*) addstoreitem(pmesh->vertexstore, &dummyvertex);
  memcpy(newvertex->coord, pvertex, sizeof(vertex3f));
  newvertex->pfirstline = NULL;
  newvertex->flag =0;

  next = (vertexindexnode*) malloc(sizeof(vertexindexnode));
    // this sets the pointers of next to NULL
  memset(next, 0, sizeof(vertexindexnode));
  next->item = newvertex;
  if (NULL != nearest)
    nearest->offsprings[index] = next;
  else
    pmesh->vertexindexroot = next;
debfuncend("VertexPointer end 1\n");

  return newvertex;
}


  /* finds the triangles neighbour on the angles[i] angles[j] side,
  sets the pointer triangle.neighbours on it, and sets the neighbours
  pointer on itself as well */
int FindNeighbourTriangleAndAdd (mesh2 *pmesh, mtriangle *ptriangle,
      int i, int j)
{
  char index;
  int itmp;
  mvertex *pvertexi, *pvertexj;
//  mtriangle *pneighbour = NULL, *pfirst;
  mline *pline, *plastline;
  mtriangle *ptriangletmp;

debfuncbegin("FindNeighbourTriangleAndAdd\n");

  if (i > j)
  { itmp = i; i = j; j = itmp; }

  pvertexi = ptriangle->angles[i];
  pvertexj = ptriangle->angles[j];

  pline = pvertexi->pfirstline;
  index = pvertexi->vindex;

  while ((NULL != pline) && (pvertexj != pline->pvertex[1-index]))
  {
/*
printf("(%f %f %f)  ",
 pline->pvertex[1-index]->coord[0],
 pline->pvertex[1-index]->coord[1],
 pline->pvertex[1-index]->coord[2]);
*/
    plastline = pline;
    pline = plastline->pnextline[index];
    index = plastline->vindex[index];
  }

  if(NULL == pline)
  {
    NewMeshLine(pmesh, ptriangle, i, j);
debfuncend("FindNeighbourTriangleAndAdd end1\n");
    return 1;
  }


    /* pline->ptriangle[0] is filled by the creating NewMeshLine() */
  ptriangletmp = pline->ptriangle[0];

  pline->ptriangle[1] = ptriangle;
  index = ((i == 0) ? ((j == 1) ? 0 : 1) : 2);
  pline->trside[1] = index;
  ptriangle->neighbours[index] = ptriangletmp;
  ptriangle->pline[index] = pline;

/*  for(itmp = 0; itmp < 3; itmp++)
  {
    if(pvertexi == ptriangletmp->angles[itmp])
      i = itmp;
    else if(pvertexj == ptriangletmp->angles[itmp])
      j = itmp;
  }

  if (i > j)
  { itmp = i; i = j; j = itmp; }

  index = ((i == 0) ? ((j == 1) ? 0 : 1) : 2); */
  index = pline->trside[0];
  ptriangletmp->neighbours[index] = ptriangle;

debfuncend("FindNeighbourTriangleAndAdd end0\n");

  return 0;
}

  /* applies 'mask' on every triangle's flag in 'pmesh' with bitwise and */
int AllTriangleFlagAndMask(mesh2 *pmesh, char mask)
{
  int i, imax;
  mtriangle *ptriangle;
  store *pstore;

  pstore = pmesh->trianglestore;
  imax = getstoreitemnum(*pstore);

  for (i = 0; i < imax; i++)
  {
    ptriangle = getstoreitemaddress(pstore, i);
    ptriangle->flag &= mask;
  }
  return 0;
}

int SetTriangleFlag(mesh2 *pmesh, int trindex, char mask)
{
  mtriangle *ptriangle;
  ptriangle = getstoreitemaddress(pmesh->trianglestore, trindex);
  if (NULL != ptriangle)
    ptriangle->flag |= mask;
}

int ResetTriangleFlag(mesh2 *pmesh, int trindex, char mask)
{
  mtriangle *ptriangle;
  ptriangle = getstoreitemaddress(pmesh->trianglestore, trindex);
  if (NULL != ptriangle)
    ptriangle->flag &= ~mask;
}


int SubDivide(mesh2 *pmeshin, mesh2 *pmeshout, float minlength, float w)
{
  int trnum, i, j, linenum, itmp;
  vertex3f *pnewvertex, vertextmp;
  char *vertexflag;
  mvertex *pmvertex;
  mline *pline;
  mtriangle *pmtriangle;
  triangle3f triangletmp;
  float ftmp;

  trnum = GetMeshTriangleNum(pmeshin);
  linenum = GetMeshLineNum(pmeshin);

  pnewvertex = malloc(trnum*3*sizeof(vertex3f));
  vertexflag = malloc(trnum*3);
  memset(vertexflag, 0, trnum*3);
#ifdef COORDTEST
glBegin(GL_LINES);
glColor3f(1,0.6,0.4);
#endif

  for(i = 0; i < linenum; i++)
  {
    pline = getstoreitemaddress(pmeshin->linestore, i);
    if (distance3f(pline->pvertex[0]->coord, pline->pvertex[1]->coord) >
      minlength)
    {
      CountDivisionVertex(pmeshin, pline, vertextmp, w);
      itmp = pline->ptriangle[0]->trindex * 3 + pline->trside[0];
      memcpy(pnewvertex[itmp], vertextmp, sizeof(vertex3f));
      vertexflag[itmp] |= TOUCHED;

      if(NULL != pline->ptriangle[1])
      {
        itmp = pline->ptriangle[1]->trindex * 3 + pline->trside[1];
        memcpy(pnewvertex[itmp], vertextmp, sizeof(vertex3f));
        vertexflag[itmp] |= TOUCHED;
      }
#ifdef COORDTEST
      else
      {
        glVertex3fv(pline->pvertex[0]->coord);
        glVertex3fv(pline->pvertex[1]->coord);
      }
#endif
      pmvertex = VertexPointer(pmeshout, vertextmp);
      pmvertex->flag |= SUBDIVVERTEX;
    }
    else // if pline length < minlength
    {
    }
  }
#ifdef COORDTEST
glEnd();
#endif

  for(i = 0; i < trnum; i++)
  {
    GetMeshTriangle3f(pmeshin, i, triangletmp);
    itmp = 3*i;
    AddSubDivTriangles(pmeshout, triangletmp, &pnewvertex[itmp],
      &vertexflag[itmp]);
//    AddTriangleToMesh(pmeshout, triangletmp);
  }

    /* move old vertexes so subdivision becomes approximating, not
    interpolating */
/*  itmp = GetMeshVertexNum(pmeshout);
  ftmp = w;
  for (i = 0; i < itmp; i++)
  {
    pmvertex = getstoreitemaddress(pmeshout->vertexstore, i);
    if (! (pmvertex->flag & SUBDIVVERTEX))
    {
      VertexAvarage(pmvertex, ftmp);
    }
  }
    */
    /* recount normals, since some vertexes are moved */
/*  itmp = GetMeshTriangleNum(pmeshout);
  for (i = 0; i < itmp; i++)
  {
    pmtriangle = getstoreitemaddress(pmeshout->trianglestore, i);
    memcpy(triangletmp[0], pmtriangle->angles[0]->coord, sizeof(vertex3f));
    memcpy(triangletmp[1], pmtriangle->angles[1]->coord, sizeof(vertex3f));
    memcpy(triangletmp[2], pmtriangle->angles[2]->coord, sizeof(vertex3f));
    GetNormal3f(triangletmp, pmtriangle->normal);
  }
*/
debfuncend("SubDivide end\n");
  return 0;
}

int GetFartherOrMid(mesh2 *pmesh, mtriangle *ptriangle, int i, vertex3f vertex)
{
  int a, b, j;
  mtriangle *ptriangle2;
  vertex3f vertextmp;

debfuncbegin("GetFartherOrMid\n");
  if(NULL != ptriangle->neighbours[i])
  {
    ptriangle2 = ptriangle->neighbours[i];
    if (ptriangle->pline[i]->ptriangle[0] == ptriangle2)
      j = 2 - ptriangle->pline[i]->trside[0];
    else
      j = 2 - ptriangle->pline[i]->trside[1];

    paOuterMatrixVertex3f(pmesh->transformmatrix,
      ptriangle2->angles[j]->coord, vertex);
  }
    /* no neighbour triangle on that side, use the midpoint of the side */
  else
  {
//printf("*\n");
    a = (i == 2 ? 2 : 0);
    b = (i == 1 ? 2 : 1);
    paOuterMatrixVertex3f(pmesh->transformmatrix,
      ptriangle->angles[a]->coord, vertex);
    paOuterMatrixVertex3f(pmesh->transformmatrix,
      ptriangle->angles[b]->coord, vertextmp);

    vertexadd3f(vertex, vertextmp);
    vertexdiv3f(vertex, 2);
  }
debfuncend("GetFartherOrMid end\n");

  return 0;
}

int CountDivisionVertex(mesh2 *pmesh, mline *pline, vertex3f vertex, float w)
{
  int i, j;
  float weight;
  vertex3f v0[2], v1[2], v2[4], vertextmp;
  mtriangle *ptriangle;
debfuncbegin("CountDivisionVertex\n");

  paOuterMatrixVertex3f(pmesh->transformmatrix,
    pline->pvertex[0]->coord, v0[0]);
  paOuterMatrixVertex3f(pmesh->transformmatrix,
    pline->pvertex[1]->coord, v0[1]);

  memcpy(vertex, v0[0], sizeof(vertex3f));
  vertexadd3f(vertex, v0[1]);
  vertexdiv3f(vertex, 2);

    /* it's so easy to get the index of the far angle of pline->ptriangle[0] */
  if(pline->ptriangle[1] != NULL)
  {
    i = 2 - pline->trside[0];
    ptriangle = pline->ptriangle[0];
    paOuterMatrixVertex3f(pmesh->transformmatrix,
      ptriangle->angles[i]->coord, v1[0]);

    i = (pline->trside[0] == 0 ? 1 : 0);
    GetFartherOrMid(pmesh, ptriangle, i, v2[0]);

    i = (pline->trside[0] == 0 ? 2 : (pline->trside[0] == 1 ? 2 : 1));
    GetFartherOrMid(pmesh, ptriangle, i, v2[1]);

    /* other triangle of pline, it's not sure whether it exists */
    ptriangle = pline->ptriangle[1];

    i = 2 - pline->trside[1];
    paOuterMatrixVertex3f(pmesh->transformmatrix,
      ptriangle->angles[i]->coord, v1[1]);

    i = (pline->trside[1] == 0 ? 1 : 0);
    GetFartherOrMid(pmesh, ptriangle, i, v2[2]);

    i = (pline->trside[1] == 0 ? 2 : (pline->trside[1] == 1 ? 2 : 1));
    GetFartherOrMid(pmesh, ptriangle, i, v2[3]);
  }
    /* if there is only one triangle on pline */
  else
  {
    return 1;
  }
/*printf("v0[0] (%f %f %f) v0[1] (%f %f %f)\n",
  v0[0][0], v0[0][1], v0[0][2], v0[1][0], v0[1][1], v0[1][2]);
printf("v1[0] (%f %f %f) v1[1] (%f %f %f)\n",
  v1[0][0], v1[0][1], v1[0][2], v1[1][0], v1[1][1], v1[1][2]);
printf("v2[0] (%f %f %f) v2[1] (%f %f %f)\n",
  v2[0][0], v2[0][1], v2[0][2], v2[1][0], v2[1][1], v2[1][2]);
printf("v2[2] (%f %f %f) v2[3] (%f %f %f)\n\n",
  v2[2][0], v2[2][1], v2[2][2], v2[3][0], v2[3][1], v2[3][2]);
*/
    /* now, move the vertex */

  memcpy(vertextmp, v1[0], sizeof(vertex3f));
  vertexadd3f(vertextmp, v1[1]);
  weight = 1.0/8.0 + 2.0*w;
  vertexmul3f(vertextmp, weight);
  vertexadd3f(vertex, vertextmp);

  memcpy(vertextmp, v2[0], sizeof(vertex3f));
  vertexadd3f(vertextmp, v2[1]);
  vertexadd3f(vertextmp, v2[2]);
  vertexadd3f(vertextmp, v2[3]);
  weight = -1.0/16.0 - w;
  vertexmul3f(vertextmp, weight);
  vertexadd3f(vertex, vertextmp);

debfuncend("CountDivisionVertex end\n");
  return 0;
}


int AddSubDivTriangles(mesh2 *pmesh, triangle3f triangle, vertex3f *pdivvertex,
      char *pvflag)
{
  int i, j, k, a, b, u, v, w, x, y, z, num;
  triangle3f triangletmp;

debfuncbegin("AddSubDivTriangles\n");
  for (i = 0, num = 0; i < 3; i++)
  {
    if(pvflag[i] & TOUCHED)
    {
      if (num == 0)
        a = i;
      else
        b = i;
      num ++;
    }
  }

  switch(num)
  {
    case 1:
        // 'a' is the index of the only divided side
      i = 2 - a;
      j = (a == 0 ? 0 : (a == 1 ? 2 : 1));

      memcpy(triangletmp[0],pdivvertex[a], sizeof(vertex3f));
      memcpy(triangletmp[1],triangle[i], sizeof(vertex3f));
      memcpy(triangletmp[2],triangle[j], sizeof(vertex3f));
      AddTriangleToMesh(pmesh, triangletmp);

      j = (a == 0 ? 1 : (a == 1 ? 0 : 2));

      memcpy(triangletmp[0],pdivvertex[a], sizeof(vertex3f));
      memcpy(triangletmp[1],triangle[j], sizeof(vertex3f));
      memcpy(triangletmp[2],triangle[i], sizeof(vertex3f));
      AddTriangleToMesh(pmesh, triangletmp);
      break;
    case 2:
    /* 'a' and 'b' are the indexes of the divided sides, a < b */
      i = (a == 0 ? (b == 1 ? 0 : 1) : 2);
      j = (a == 0 ? (b == 1 ? 1 : 0) : 0);
      k = (a == 0 ? (b == 1 ? 2 : 2) : 1);

      if ((a == 0) && (b == 2))
      {
        memcpy(triangletmp[0],pdivvertex[b], sizeof(vertex3f));
        memcpy(triangletmp[1],pdivvertex[a], sizeof(vertex3f));
      }
      else
      {
        memcpy(triangletmp[0],pdivvertex[a], sizeof(vertex3f));
        memcpy(triangletmp[1],pdivvertex[b], sizeof(vertex3f));
      }
      memcpy(triangletmp[2],triangle[i], sizeof(vertex3f));
      AddTriangleToMesh(pmesh, triangletmp);

      if(distance3f(triangle[j],pdivvertex[b]) <
        distance3f(triangle[k],pdivvertex[a]))
      {
        if ((a == 0) && (b == 1))
          u = 1, v = 0, w = 1, x = 1, y = 1,z = 2;
        else if ((a == 0) && (b == 2))
          u = 0, v = 2, w = 0, x = 2, y = 2,z = 0;
        else
          u = 2, v = 1, w = 0, x = 2, y = 0,z = 1;
      }
      else
      {
        if ((a == 0) && (b == 1))
          u = 1, v = 0, w = 2, x = 0, y = 1,z = 2;
        else if ((a == 0) && (b == 2))
          u = 0, v = 2, w = 2, x = 0, y = 2,z = 0;
        else
          u = 2, v = 1, w = 1, x = 1, y = 0,z = 1;
      }

      memcpy(triangletmp[0],pdivvertex[u], sizeof(vertex3f));
      memcpy(triangletmp[1],pdivvertex[v], sizeof(vertex3f));
      memcpy(triangletmp[2],triangle[w], sizeof(vertex3f));
      AddTriangleToMesh(pmesh, triangletmp);

      memcpy(triangletmp[0],pdivvertex[x], sizeof(vertex3f));
      memcpy(triangletmp[1],triangle[y], sizeof(vertex3f));
      memcpy(triangletmp[2],triangle[z], sizeof(vertex3f));
      AddTriangleToMesh(pmesh, triangletmp);
      break;
    case 3:
      memcpy(triangletmp[0],pdivvertex[0], sizeof(vertex3f));
      memcpy(triangletmp[1],pdivvertex[1], sizeof(vertex3f));
      memcpy(triangletmp[2],triangle[0], sizeof(vertex3f));
      AddTriangleToMesh(pmesh, triangletmp);

      memcpy(triangletmp[0],pdivvertex[2], sizeof(vertex3f));
      memcpy(triangletmp[1],pdivvertex[0], sizeof(vertex3f));
      memcpy(triangletmp[2],triangle[1], sizeof(vertex3f));
      AddTriangleToMesh(pmesh, triangletmp);

      memcpy(triangletmp[0],pdivvertex[1], sizeof(vertex3f));
      memcpy(triangletmp[1],pdivvertex[2], sizeof(vertex3f));
      memcpy(triangletmp[2],triangle[2], sizeof(vertex3f));
      AddTriangleToMesh(pmesh, triangletmp);

      memcpy(triangletmp[0],pdivvertex[0], sizeof(vertex3f));
      memcpy(triangletmp[1],pdivvertex[2], sizeof(vertex3f));
      memcpy(triangletmp[2],pdivvertex[1], sizeof(vertex3f));
      AddTriangleToMesh(pmesh, triangletmp);

      break;
    default:
      AddTriangleToMesh(pmesh, triangle);
      break;
  } // end switch(num)

debfuncend("AddSubDivTriangles end\n");
  return 0;
}

int ismember(void *in, int num, int size, void *this)
{
  int i;
  char *pinc, *pthisc;
  pinc = (char*) in;
  pthisc = (char*) this;

  for(i = 0; i < num; i++)
  {
    if(0 == memcmp(&pinc[i*size], pthisc, size))
      return 1;
  }
  return 0;
}

int VertexAvarage(mvertex *pmvertex, float weight)
{
  int i, j, k;
  char vindex, vindextmp;
  mvertex *pnmvertex;
  mvertex **ppmvertex;
  mline *pline;
  vertex3f vertextmp = {0.0, 0.0, 0.0};
//  stack st;
//return 0;
//  initstack(&st);
  i = 0;
  pline = pmvertex->pfirstline;
  vindex = pmvertex->vindex;
  while(pline != NULL)
  {
//    push(&st, &pline->pvertex[1-vindex], sizeof(mvertex*));
    vertexadd3f(vertextmp, pline->pvertex[1-vindex]->coord);
    vertexsub3f(vertextmp, pmvertex->coord);
    vindextmp = vindex;
    vindex = pline->vindex[vindextmp];
    pline = pline->pnextline[vindextmp];
    i++;
  }

  if(i)
  {
    vertexmul3f(vertextmp, 1.0/(i*4) + weight*4/i);
    vertexadd3f(pmvertex->coord, vertextmp);

/*    memset(vertextmp, 0, sizeof(vertex3f));
    ppmvertex = getstackunsafe(&st, &i);
    i /= sizeof(mvertex*);
    k = 0;
    for(j = 0; j < i; j++)
    {
      pline = ppmvertex[j]->pfirstline;
      vindex = ppmvertex[j]->vindex;
      while(pline != NULL)
      {
        pnmvertex = pline->pvertex[1-vindex];
        if (( pnmvertex != pmvertex) &&
          !ismember(ppmvertex, i, sizeof(mvertex*), &pnmvertex))
        {
          vertexadd3f(vertextmp, pnmvertex->coord);
          vertexsub3f(vertextmp, pmvertex->coord);
          k++;
        }
        vindextmp = vindex;
        vindex = pline->vindex[vindextmp];
        pline = pline->pnextline[vindextmp];
      }
    }
    if (k)
    {
      vertexmul3f(vertextmp, -1.0/(i*4) - weight*4/i);
      vertexadd3f(pmvertex->coord, vertextmp);
    }*/
//    vertexmul3f(pmvertex->coord, 1 - weight);
  }

//  freestack(&st);
  return 0;
}


int MeshToRib(mesh2 *pmesh, FILE* fil, const char* linebegin)
{
  int i, trnum;
  triangle3f triangletmp;

  trnum = GetMeshTriangleNum(pmesh);
  for(i = 0; i < trnum ; i++)
  {
    GetMeshTriangle3f(pmesh, i, triangletmp);
    fprintf (fil, "%sPolygon \"P\" [%f %f %f  %f %f %f  %f %f %f]\n",
      linebegin,
      triangletmp[0][0], triangletmp[0][1], triangletmp[0][2],
      triangletmp[1][0], triangletmp[1][1], triangletmp[1][2],
      triangletmp[2][0], triangletmp[2][1], triangletmp[2][2]);
  }

  return 0;
}

#ifdef WITHGL
#include <GL/glut.h>
int DrawMeshGL(mesh2 *pmesh, int type)
{
  int i, maxi;
  triangle3f triangletmp;
  vector3f vectortmp;
  line3f pline;
  mline *pmline;

  switch (type)
  {
  case DRAW_SOLID:
    maxi=getstoreitemnum(*pmesh->trianglestore);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glBegin(GL_TRIANGLES);
    for (i = 0; i < maxi; i++)
    {
      GetMeshTriangleAndNormal3f(pmesh, i, triangletmp, vectortmp);
      glNormal3fv(vectortmp);
      glVertex3fv(triangletmp[0]);
      glVertex3fv(triangletmp[1]);
      glVertex3fv(triangletmp[2]);
    }
    glEnd();
    break;
  case DRAW_WIRE:
    maxi=getstoreitemnum(*pmesh->linestore);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    glBegin(GL_LINES);
    glColor3fv(pmesh->color);
    for (i = 0; i < maxi; i++)
    {
      GetMeshLine3f(pmesh, i, pline);
      glVertex3fv(pline[0]);
      glVertex3fv(pline[1]);
    }
    glEnd();
    break;
  case DRAW_DEBUG_WIRE:
    maxi=getstoreitemnum(*pmesh->linestore);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    glBegin(GL_LINES);
    for (i = 0; i < maxi; i++)
    {
      GetMeshLine3f(pmesh, i, pline);
      pmline = getstoreitemaddress(pmesh->linestore, i);
      if (NULL == pmline->ptriangle[1])
        glColor3f(1, 0, 1);
      else
        glColor3fv(pmesh->color);
      glVertex3fv(pline[0]);
      glVertex3fv(pline[1]);
    }
    glEnd();
    break;
  }
  return 0;
}


int DrawLineAndNormalGL (sectline *pline)
{
  vertex3f vertextmp;
  vector3f vectortmp;
  float r;

 // r = (float) rand()/ 5.0 / RAND_MAX - 0.1;
 if(distance3f(pline->endpoint[0],pline->endpoint[1]) > 0.1)
   return 1;


glColor3f(1,0,1);
  memcpy(vertextmp, pline->endpoint[0], sizeof(vertex3f));
//  vertextmp[0] += r;
  glVertex3fv(vertextmp);
  memcpy(vertextmp, pline->endpoint[1], sizeof(vertex3f));
//  vertextmp[0] += r;
  glVertex3fv(vertextmp);
  vertexadd3f (vertextmp, pline->endpoint[0]);
  vertextmp[0] /= 2;
  vertextmp[1] /= 2;
  vertextmp[2] /= 2;
  glColor3f(0,1,0);
  glVertex3fv(vertextmp);
  memcpy(vectortmp, pline->normal, sizeof(vector3f));
  vectortmp[0] /= 12;
  vectortmp[1] /= 12;
  vectortmp[2] /= 12;
  vertexadd3f(vertextmp, vectortmp);
  glVertex3fv(vertextmp);
}

#endif // WITHGL
