/*
myio.c
file handling functions
Prim Andras 2002
*/

#include "myio.h"

int debugindent = 0;
void debugprintf(char *string, int indent)
{
  int i;
  if (indent < 0)
    debugindent += indent;
  for (i = 0; i < debugindent; i++)
    putchar('-');
  printf(string);
  if (indent > 0)
    debugindent += indent;
}


 /* skips lines with only blank characters and comments
    comment is a line which looks like <blanks>'#'<anything>'\n' */
void fskipNLsorcomments(FILE *fil)
{
  long linebegin;
  int ch, stillgoing=1, comment = 0;

debinfo("fskipNLSorcomments\n");
  linebegin = ftell(fil);

  do
  {
    ch = fgetc(fil);
    switch(ch)
    {
      case CR:
      case LF:
        linebegin = ftell(fil);
        comment = 0;
        break;
      case '#':
        comment = 1;
        break;
      case ' ':
      case '\t':
        break;
      case EOF:
        stillgoing = 0;
      default:
        if (!comment)
        {
          stillgoing = 0;
          fseek (fil, linebegin, SEEK_SET);
        }
        break;
    }
  }while(stillgoing);
debinfo("fskipNLSorcomments end\n");
} // fskipNLsorcomments();


 /* skips till the next newline */
void fskipline(FILE *fil)
{
  int ch;
  do
  {
    ch = fgetc(fil);
  }while ((LF != ch) && (EOF != ch));
} // fskipline()


 /* reads a string beginning with " character, ending with " or NEWLINE from
    'fil' returns the string in a newly allocated buffer
    returns NULL if EOF reached without reading '"' */
char *freadstring(FILE *fil)
{
  char *stmp;
  long stringbegin, length;
  int ch;

  do
  {
    if (EOF == (ch = fgetc(fil)))
      return NULL;
  }while ('"' != ch);

  stringbegin = ftell(fil);

  do
  {
    ch = fgetc (fil);
  }while ( ('"' != ch) && (LF != ch) && (EOF != ch) && (CR != ch));

  length = ftell(fil) - stringbegin - 1;
  stmp = (char*) malloc(length + 1);

  fseek (fil, stringbegin, SEEK_SET);
  fread (stmp, 1, length, fil);
  stmp[length] = 0;

  return stmp;
}  // freadstring()

int freadfloat(FILE *fil, float *num, const char *sep)
{
  float dec = 0.1, val = 0;
  int ch, gottowork = 1, decpoint = 0, pol = 1;

    // characters before the number
  do {
    ch = fgetc(fil);
    if (ch == EOF)
      return 2;
    else if (NULL != strchr(sep, ch))
      return 1;

    if(!isdigit(ch))
    {
        // if ch == '.', than we shall not change pol
      if ('.' == ch)
        decpoint = 1;
      else
      {
        decpoint = 0;
        if ('-' == ch)
          pol = -1;
        else
          pol = 1;
      }
    }
  }while(!isdigit(ch));

  do {
    if (decpoint)
    {
      val += dec * (float)(ch - '0');
      dec /= 10;
    }
    else
      val = (val * 10) + (float)(ch - '0');

    ch = fgetc(fil);

    if (ch == '.')
    {
      decpoint = 1;
        // skip decimal point
      ch = fgetc(fil);
    }
    if (!isdigit(ch))
      gottowork = 0;
  }while (gottowork);

  *num = val * (float)pol;
  if (EOF != ch)
    fseek(fil, -1, SEEK_CUR);
  return 0;
}

