/*
mystring.c
own string handling functions
Prim Andras 2002
*/

#include "mystring.h"
#include <stdlib.h>
#include <string.h>

#ifdef DEBUG
#define debinfo(a) printf(a)
#else
#define debinfo(a)
#endif


void InitDS(dynastring *ds)
{
  ds->first = NULL;
  ds->act = NULL;
  ds->i = 0;
  ds->length = 0;
}

  // copies stringin into ds
void AddStringToDS(dynastring *ds, char *stringin, int lengthrem)
{
  int i, copybytenum;
  dynalink *act, *tmp;

/*  debinfo("string\"");
  debinfo(stringin);
  debinfo("\"\n");
*/
    // first link shall be created
  if (NULL == ds->act)
  {
    act = malloc(sizeof(dynalink));
    act->next = NULL;
    act->length = 0;
    ds->first = act;
    i = 0;
  }
  else
  {
    act = ds->act;
    i = ds->i;
  }

  while (lengthrem > 0)
  {
    copybytenum = (DYNABUFLENGTH - i) < lengthrem ?
                  (DYNABUFLENGTH - i) : lengthrem;
    memcpy(act->buf + i, stringin, copybytenum);

    lengthrem -= copybytenum;
    stringin += copybytenum;
/*  debinfo("stringtrunc\"");
  debinfo(stringin);
  debinfo("\"\n");
*/
    act->length += copybytenum;
     /* probably a "ds->length += lengthrem" before the while would be a bit
        better */
    ds->length += copybytenum;

      // new link
    if( (i + copybytenum) >= DYNABUFLENGTH )
    {
      i = 0;
      tmp = malloc(sizeof(dynalink));
      tmp->next = NULL;
      tmp->length = 0;
      act->next = tmp;
      act = tmp;
    }
    else
    {
      i += copybytenum;
    }
  }

  ds->act = act;
  ds->i = i;
}


  /* adds character ch to ds
  */
void AddCharToDS(dynastring *ds, char ch)
{
  int i;
  dynalink *act, *tmp;

    // first link shall be created
  if (NULL == ds->act)
  {
    act = malloc(sizeof(dynalink));
    act->next = NULL;
    act->length = 0;
    ds->first = act;
    i = 0;
  }
  else
  {
    act = ds->act;
    i = ds->i;
  }

  act->buf[i++] = ch;
  act->length++;
  ds->length++;

  if (i >= DYNABUFLENGTH)
  {
      i = 0;
      tmp = malloc(sizeof(dynalink));
      tmp->next = NULL;
      tmp->length = 0;
      act->next = tmp;
      act = tmp;
  }

  ds->act = act;
  ds->i = i;
} // AddCharToDS()


  /* adds float ch to ds
  */
void AddFloatToDS(dynastring *ds, float fl)
{
#ifdef FLOATTEXT
  char stmp[FLOATLENGTH];
  sprintf(stmp, "%f", fl);
  AddStringToDS(ds, stmp, strlen(stmp));
#else
  char stmp[sizeof(float)+2];
  stmp[0] = FLOATESCAPE;
  memcpy(&(stmp[1]), &fl, sizeof(float));
  AddStringToDS(ds, stmp, sizeof(float) + 1);
#endif
} // AddFloatToDS()


  // returns the string stored in ds in a newly allocated buffer
char *GetStringFromDS(dynastring *ds, int *length)
{
  char *retstring, *actstring;
  dynalink *act;

  retstring = malloc(ds->length + 1);
  actstring = retstring;
  act = ds->first;

  while(act != NULL)
  {
    memcpy(actstring, act->buf, act->length);
    actstring += act->length;
    act = act->next;
  }
  *actstring = 0;

  *length = ds->length;
  return retstring;
}


  // frees memory allocated by the chain of ds, and initialises ds
void FreeDS(dynastring *ds)
{
  dynalink *act, *last;

  act = ds->first;

  while (NULL != act)
  {
    last = act;
    act = act->next;
    free(last);
  }
    // from this point: InitDS(ds)
  ds->first = NULL;
  ds->act = NULL;
  ds->i = 0;
  ds->length = 0;
}


 /* returns the length of shortstr, if longstr begins with shortstr
 otherwise returns 0 */
int strbegin(char *longstr, char *shortstr)
{
  int i;

  for(i = 0; ; i++)
  {
    if(0 == shortstr[i])
      return i;
    if(longstr[i] != shortstr[i])
      return 0;
  }
}
