/*
pagl.c
OpenGL like matrix functions
Prim Andras 2002
*/
#include "pagl.h"
#include "pastuff.h"
#include <math.h>
#include <stdlib.h>

matrix4x4f paMatrix;
/*matrix4x4f paIdentity = {{1.0, 0.0, 0.0, 0.0},
                         (0.0, 1.0, 0.0, 0.0},
                         {0.0, 0.0, 1.0, 0.0},
                         {0.0, 0.0, 0.0, 1.0}}
*/
matrix4x4f paIdentity = {1.0, 0.0, 0.0, 0.0,
                         0.0, 1.0, 0.0, 0.0,
                         0.0, 0.0, 1.0, 0.0,
                         0.0, 0.0, 0.0, 1.0};


  /* the functions below try to act like the ones in OpenGL */
  /* Loads identity matrix into paMatrix */
void paLoadIdentity (void)
{
  memcpy(paMatrix, paIdentity, sizeof(matrix4x4f));
}


  /* Loads the matrix */
void paLoadMatrixf (matrix4x4f matrix)
{
  memcpy(paMatrix, matrix, sizeof(matrix4x4f));
}


  /* adds a rotation to paMatrix */
void paRotatef (float fi, float xin, float yin, float zin)
{
  matrix4x4f matrix;
  float x, y, z, lengthin, cosfi, cosfi2, sinfi;

  lengthin = sqrt(xin*xin + yin*yin + zin * zin);
    /* given a 0 long vector to rotate around: don't like it */
  if (nullf(lengthin))
    return;
  x = xin / lengthin; y = yin / lengthin; z = zin / lengthin;
  memset (matrix, 0, sizeof(matrix4x4f));
  matrix[3][3] = 1.0;

    // convert fi to radian
  fi = fi * M_PI / 180;
  cosfi = cos(fi);
  cosfi2 = 1 - cosfi;
  sinfi = sin(fi);
  matrix[0][0] = cosfi + (cosfi2)*x*x;
  matrix[0][1] = (cosfi2)*x*y + sinfi*z;
  matrix[0][2] = (cosfi2)*z*x - sinfi*y;
  matrix[1][0] = (cosfi2)*x*y - sinfi*z;
  matrix[1][1] = cosfi + (cosfi2)*y*y;
  matrix[1][2] = (cosfi2)*z*y + sinfi*x;
  matrix[2][0] = (cosfi2)*x*z + sinfi*y;
  matrix[2][1] = (cosfi2)*y*z - sinfi*x;
  matrix[2][2] = cosfi + (cosfi2)*z*z;

  paMultMatrixf(matrix);
}


  /* adds a translation to paMatrix */
void paTranslatef (float x, float y, float z)
{
  matrix4x4f matrix;
  if (nullf(x) && nullf(y) && nullf(z))
    return;
  memset (matrix, 0, sizeof(matrix4x4f));
  matrix[0][0] = 1.0;
  matrix[1][1] = 1.0;
  matrix[2][2] = 1.0;
  matrix[3][3] = 1.0;
  matrix[3][0] = x;
  matrix[3][1] = y;
  matrix[3][2] = z;

  paMultMatrixf(matrix);
}


  /* multiplies paMatrix with matrix */
void paMultMatrixf (matrix4x4f matrix)
{
  matrix4x4f matrixtmp;
  int i, j;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      matrixtmp[i][j] = paMatrix[0][j]*matrix[i][0] +
                        paMatrix[1][j]*matrix[i][1] +
                        paMatrix[2][j]*matrix[i][2] +
                        paMatrix[3][j]*matrix[i][3];

  memcpy(paMatrix, matrixtmp, sizeof(matrix4x4f));
}


  /* returns a pointer to paMatrix */
matrix4x4f* paGetMatrixf (void)
{
  return &paMatrix;
}


  /* multiplies vertexin with paMatrix -- puts the result into vertexout */
void paMatrixVertex3f (vertex3f vertexin, vertex3f vertexout)
{
  int i;
  float x, y, z;

    /* not really for speed-up reason, but it handles the case, when
    vertexin is the same address as vertexout */
  x = vertexin[0];
  y = vertexin[1];
  z = vertexin[2];
  for(i = 0; i < 3; i ++)
    vertexout[i] = paMatrix[0][i]*x + paMatrix[1][i]*y +
                      paMatrix[2][i]*z + paMatrix[3][i];
}

  /* multiplies vectorin with paMatrix -- puts the result into vectorout
  --- same as paMetrixVertex3f, except that translation takes no effect here */
void paMatrixVector3f (vector3f vectorin, vector3f vectorout)
{
  int i;
  float x, y, z;

  x = vectorin[0];
  y = vectorin[1];
  z = vectorin[2];

  for(i = 0; i < 3; i ++)
    vectorout[i] = paMatrix[0][i]*x + paMatrix[1][i]*y +
                      paMatrix[2][i]*z;
}

stack pastack = {0, 0, NULL};

void paPushMatrix(void)
{
  push(&pastack, &paMatrix, sizeof(matrix4x4f));
}


void paPopMatrix(void)
{
  if (!pastack.size)
    return;
  pop(&pastack, &paMatrix, sizeof(matrix4x4f));
}


// Outer versions


void paOuterLoadIdentity (matrix4x4f matrixdest)
{
  memcpy(matrixdest, paIdentity, sizeof(matrix4x4f));
}

void paOuterRotatef (matrix4x4f matrixdest, float fi, float xin, float yin,
      float zin)
{
  matrix4x4f matrix;
  float x, y, z, lengthin, cosfi, cosfi2, sinfi;

  lengthin = sqrt(xin*xin + yin*yin + zin * zin);
    /* given a 0 long vector to rotate around: don't lik it */
  if (nullf(lengthin))
    return;
  x = xin / lengthin; y = yin / lengthin; z = zin / lengthin;
  memset (matrix, 0, sizeof(matrix4x4f));
  matrix[3][3] = 1.0;

    // convert fi to radian
  fi = fi * M_PI / 180;
  cosfi = cos(fi);
  cosfi2 = 1 - cosfi;
  sinfi = sin(fi);
  matrix[0][0] = cosfi + (cosfi2)*x*x;
  matrix[0][1] = (cosfi2)*x*y + sinfi*z;
  matrix[0][2] = (cosfi2)*z*x - sinfi*y;
  matrix[1][0] = (cosfi2)*x*y - sinfi*z;
  matrix[1][1] = cosfi + (cosfi2)*y*y;
  matrix[1][2] = (cosfi2)*z*y + sinfi*x;
  matrix[2][0] = (cosfi2)*x*z + sinfi*y;
  matrix[2][1] = (cosfi2)*y*z - sinfi*x;
  matrix[2][2] = cosfi + (cosfi2)*z*z;

  paOuterMultMatrixf(matrixdest, matrix);
}


void paOuterTranslatef (matrix4x4f matrixdest, float x, float y, float z)
{
  matrix4x4f matrix;
  if (nullf(x) && nullf(y) && nullf(z))
    return;
  memset (matrix, 0, sizeof(matrix4x4f));
  matrix[0][0] = 1.0;
  matrix[1][1] = 1.0;
  matrix[2][2] = 1.0;
  matrix[3][3] = 1.0;
  matrix[3][0] = x;
  matrix[3][1] = y;
  matrix[3][2] = z;

  paOuterMultMatrixf(matrixdest, matrix);
}


void paOuterMultMatrixf (matrix4x4f matrixdest, matrix4x4f matrix)
{
  matrix4x4f matrixtmp;
  int i, j;

  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      matrixtmp[i][j] = matrixdest[0][j]*matrix[i][0] +
                        matrixdest[1][j]*matrix[i][1] +
                        matrixdest[2][j]*matrix[i][2] +
                        matrixdest[3][j]*matrix[i][3];

  memcpy(matrixdest, matrixtmp, sizeof(matrix4x4f));
}

void paOuterMatrixVertex3f (matrix4x4f matrixdest, vertex3f vertexin, vertex3f
 vertexout)
{
  int i;
  float x, y, z;

  x = vertexin[0];
  y = vertexin[1];
  z = vertexin[2];

  for(i = 0; i < 3; i ++)
    vertexout[i] = matrixdest[0][i]*x + matrixdest[1][i]*y +
                      matrixdest[2][i]*z + matrixdest[3][i];
}

void paOuterMatrixVector3f (matrix4x4f matrixdest, vector3f vectorin, vector3f
 vectorout)
{
  int i;
  float x, y, z;

  x = vectorin[0];
  y = vectorin[1];
  z = vectorin[2];

  for(i = 0; i < 3; i ++)
    vectorout[i] = matrixdest[0][i]*x + matrixdest[1][i]*y +
                      matrixdest[2][i]*z;
}


