/*
pagl.h
OpenGL like matrix functions
Prim Andras 2002
*/
#ifndef PAGL_H
#define PAGL_H

#define EPSILON 0.001

#define notnullf(a) (((a) > EPSILON) || ((a) < -EPSILON))
#define nullf(a) (((a) < EPSILON) && ((a) > -EPSILON))
#define ltf(a,b) ((b) - (a) > EPSILON)
#define ltef(a,b) ((b) - (a) > -EPSILON)

#define EPSILON2 0.00001

#define notnullf2(a) (((a) > EPSILON2) || ((a) < -EPSILON2))
#define nullf2(a) (((a) < EPSILON2) && ((a) > -EPSILON2))
#define ltf2(a,b) ((b) - (a) > EPSILON2)
#define ltef2(a,b) ((b) - (a) > -EPSILON2)

typedef float vector3f [3];
typedef vector3f vertex3f;
typedef float vector4f [4];
typedef vertex3f line3f [2];
typedef vector3f triangle3f [3];

typedef float matrix4x4f [4][4];

  /* the functions below try to act like the ones in OpenGL */
  /* Loads identity matrix into paMatrix */
void paLoadIdentity (void);
  /* Loads the matrix */
void paLoadMatrixf (matrix4x4f matrix);
  /* adds a rotation to paMatrix */
void paRotatef (float deg, float x, float y, float z);
  /* adds a translation to paMatrix */
void paTranslatef (float x, float y, float z);
  /* multiplies paMatrix with matrix */
void paMultMatrixf (matrix4x4f matrix);
  /* returns a pointer to paMatrix */
matrix4x4f* paGetMatrixf (void);
  /* multiplies vertexin with paMatrix -- puts the result into vertexout */
void paMatrixVertex3f (vertex3f vertexin, vertex3f vertexout);
  /* multiplies vectorin with paMatrix -- puts the result into vectorout
  --- same as paMetrixVertex3f, except that translation takes no effect here */
void paMatrixVector3f (vector3f vectorin, vector3f vectorout);
void paPushMatrix(void);
void paPopMatrix(void);

void paOuterLoadIdentity (matrix4x4f matrixdest);
void paOuterRotatef (matrix4x4f matrixdest, float deg, float x, float y,
      float z);
void paOuterTranslatef (matrix4x4f matrixdest, float x, float y, float z);
void paOuterMultMatrixf (matrix4x4f matrixdest, matrix4x4f matrix);
void paOuterMatrixVertex3f (matrix4x4f matrixdest, vertex3f vertexin, vertex3f
 vertexout);
void paOuterMatrixVector3f (matrix4x4f matrixdest, vector3f vectorin, vector3f
 vectorout);

#endif // first inclusion of this header
