#include <stdlib.h>
#include "pastuff.h"


// stack begin ***********************************
void initstack(stack *st)
{
  st->ptr = 0;
  st->size = 0;
  st->buf = NULL;
}

void freestack(stack *st)
{
  if (NULL != st->buf)
    free(st->buf);

  st->ptr = 0;
  st->size = 0;
}

void *push(stack *st, void *this, int size)
{
  void *dest;
  if(st->ptr + size > st->size)
  {
    st->size = st->ptr + max(size, STACKBURST);
    st->buf = realloc(st->buf, st->size);
  }
  dest = &(st->buf[st->ptr]);
  memcpy(dest, this, size);
  st->ptr += size;
  return dest;
}

void pop(stack *st, void *this, int size)
{
  if (size > st->ptr)
  {
    memset(this, 0, size);
    size = st->ptr;
  }
  if (st->ptr)
  {
    st->ptr -= size;
    memcpy(this, &(st->buf[st->ptr]), size);
  }
}

void *getstack(stack *st, int *size)
{
  char *ctmp = malloc(st->ptr);
  memcpy(ctmp, st->buf, st->ptr);
  *size = st->ptr;
  return ctmp;
}

void *getstackunsafe(stack *st, int *size)
{
  *size = st->ptr;
  return st->buf;
}
// stack end *************************************


// store begin ***********************************
void initstore (store *st, int itemsize)
{
  st->itemsize = itemsize;
  st->burstsize = STOREBURST;
  st->items = NULL;
  st->burstnum = 0;
  st->actualburst = 0;
  st->index = 0;
}

void initstore2 (store *st, int itemsize, int burstsize)
{
  st->itemsize = itemsize;
  st->burstsize = burstsize;
  st->items = NULL;
  st->burstnum = 0;
  st->actualburst = 0;
  st->index = 0;
}

store *newstore (int itemsize)
{
  store *ret;
  ret = malloc (sizeof(store));
  initstore(ret, itemsize);

  return ret;
}


store *newstore2 (int itemsize, int burstsize)
{
  store *ret;
  ret = malloc (sizeof(store));
  initstore2(ret, itemsize, burstsize);

  return ret;
}

void freestaticstore (store *st)
{
  int i;
  if (NULL == st->items)
    return;
  for(i = 0; i < st->burstnum; i++)
  {
    free(st->items[i]);
  }
  free(st->items);

  st->items = NULL;
  st->burstnum = 0;
  st->actualburst = 0;
  st->index = 0;
}

void freestore (store *st)
{
  freestaticstore(st);
  free (st);
}

void *addstoreitem (store *st, void *item)
{
  void *place;
  if (st->index >= st->burstsize)
  {
    st->actualburst ++;
    st->index = 0;
  }

  if ((NULL == st->items) ||
    (st->actualburst >= st->burstnum) )
  {
      // gets the index of the new buffer
    st->actualburst = st->burstnum;
    st->burstnum++;
    st->items = realloc(st->items, st->burstnum*sizeof(void*));
    st->items[st->actualburst] = malloc(st->itemsize*st->burstsize);
  }

  place = ((char*) (st->items[st->actualburst])) + st->index*st->itemsize;

  memcpy(place, item, st->itemsize);
  st->index++;
  return place;
}

void *getstoreitemaddress (store *st, int index)
{
  int burstindex;
  if (index > getstoreitemnum(*st))
    return NULL;
  else
  {
    burstindex = index / st->burstsize;
// printf("index, burstsize, itemsize");
    index = index % st->burstsize;

    return ((char*) st->items[burstindex]) + index*st->itemsize;
  }
}

void getstoreitem (store *st, int index, void *item)
{
  void *source;
  source = getstoreitemaddress(st, index);
  memcpy(item, source, st->itemsize);
}

void *getstore (store *st)
{
  int i, itemnum;
  char *ret, *ret2;

  if ( 0 == (itemnum = getstoreitemnum(*st)))
    return NULL;

  ret2 = ret = (char*) malloc(itemnum * st->itemsize);
  for (i = 0; i < st->actualburst; i++, ret+=(st->burstsize)*(st->itemsize))
    memcpy(ret2, st->items[i], (st->burstsize)*(st->itemsize));


  memcpy(ret2, st->items[i], st->index * st->itemsize);

  return (void*) ret;
}

/*
int getstoreitemnum(store *st)
{
 return st->actualburst * st->burstsize + st->index;
}
*/
// store end *************************************
