/* turtle3d.c
graphically represent the turtle string
Prim Andras 2002
*/

#include <math.h>
#include "pastuff.h"
#include "turtle3d.h"
#include "myio.h"
#include "mystring.h"

#define MINLENGTH 0.001
#define W_DEFAULT -0.01

float basecos[BRANCHBASEPOINTS+1];
float basesin[BRANCHBASEPOINTS+1];

/* creates the mesh of a tree given with the string 'tree' in a form, that is
  easy to handle with GL_TRIANGLES
  returns number of triangles */
int CreateTreeMesh(plant *pl, vertex3f **vertexes, vector3f **normals)
{
  mesh2 *ptreemesh, *pleafmesh;
  int trnum;

  ptreemesh = NewMesh();
  pleafmesh = NewMesh();

  CreateLeafyTreeMesh(pl, ptreemesh, pleafmesh);
  trnum = GetMeshTrianglesAndNormals3f(ptreemesh, vertexes, normals);
  FreeMesh (ptreemesh);
  FreeMesh (pleafmesh);

  return trnum;
}

/* creates the tree and adds the branches to ptreemesh and the leaves
 to pleafmesh */
int CreateLeafyTreeMesh(plant *pl, mesh2 *ptreemesh, mesh2 *pleafmesh)
{
debinfo("CreateTreeMesh\n");
    // this will hold width of branch and pointer to branches
  stack st;
  branch *root;
  branch *actbranch;
  char *tree;

  float param, width = 1, curcos, cursin, deg = 2 * M_PI / BRANCHBASEPOINTS;
  float param2, param3;
  int ch, sign, i;

  tree = pl->sentence;
  root = (branch*) malloc(sizeof(branch));
  root->offspringnum = -1;
  root->offsprings = NULL;
  root->psurface = NULL;
  for (i = 1; i < BRANCHBASEPOINTS; i++)
  {
    basesin[i] = sin(i*deg);
    basecos[i] = cos(i*deg);
  }

  actbranch = root;

  initstack(&st);
  paLoadIdentity();
debinfo("CreateTreeMesh 1\n");

  while (*tree != 0)
  {
    sign = 1;
    switch(*tree)
    {
      case '[':
        paPushMatrix();
        push(&st, &width, sizeof(float));
        push(&st, &actbranch, sizeof(branch*));
        break;
      case ']':
        if(getstacklength(&st) >= (sizeof(branch*) + sizeof(float)) )
        {
          paPopMatrix();
          pop(&st, &actbranch, sizeof(branch*));
          pop(&st, &width, sizeof(float));
        }
        break;
      case 'F':
          // skips 'F'
        tree++;
        if('(' == *tree)
        {
          tree++;
          if (getfloat(&tree, &param))
            param = DEFBRANCHLENGTH;
        }
        else
        {
          param = DEFBRANCHLENGTH;
          tree--;
        }
        actbranch = NewBranch(actbranch, width, param);
        break;
      case 'L':
          // skips 'L('
        tree++;
        if (*tree == '(')
        {
          tree++;
          if (getfloat(&tree, &param))
            param = DEFLEAFPOS;
          else
            tree++;

          if (getfloat(&tree, &param2))
            param2 = DEFLEAFANGLE;
          else
            tree++;

          if (getfloat(&tree, &param3))
          {
            tree--;
            param3 = DEFLEAFSIZE;
          }
        }
        else
        {
          tree--;
          param = DEFLEAFPOS;
          param2 = DEFLEAFANGLE;
          param3 = DEFLEAFSIZE;
        }

        NewLeaf(pleafmesh, &pl->ls->leaf, actbranch,
          param, param2, param3);

        break;
      case '!':
        tree++;
        if('(' == *tree)
        {
          tree++;
          if (getfloat(&tree, &param))
            param = DEFWIDTH;
        }
        else
        {
          param = DEFWIDTH;
          tree--;
        }
        width = param;
        break;
      case '-':
        sign = -1;
      case '+':
        tree++;
        if('(' == *tree)
        {
          tree++;
          if (getfloat(&tree, &param))
            param = DEFTURNANGLE;
        }
        else
        {
          param = DEFTURNANGLE;
          tree--;
        }
        param *= sign;
        paRotatef(param, 0.0, 0.0, 1.0);
        break;
      case '^':
        sign = -1;
      case '&':
        tree++;
        if('(' == *tree)
        {
          tree++;
          if (getfloat(&tree, &param))
            param = DEFTURNANGLE;
        }
        else
        {
          param = DEFTURNANGLE;
          tree--;
        }
        param *= sign;
        paRotatef(param, 1.0, 0.0, 0.0);
        break;
      case '/':
        sign = -1;
      case '\\':
        tree++;
        if('(' == *tree)
        {
          tree++;
          if (getfloat(&tree, &param))
            param = DEFTURNANGLE;
        }
        else
        {
          param = DEFTURNANGLE;
          tree--;
        }
        param *= sign;
        paRotatef(param, 0.0, 1.0, 0.0);
        break;
    }

    tree++;
  } // while (*tree != 0)

debinfo("CreateTreeMesh 2\n");
  freestack(&st);
  AddBranchMeshAndDieRec(root, ptreemesh);
debinfo("CreateTreeMesh\n");
  return 0;
}


/* creates a new branch at the endpoint of 'father' according to paMatrix, gives
himself to 'father->offsprings' and translates paMatrix to its endpoint.
returns a pointer to the new branch */
branch *NewBranch(branch* father, float width, float length)
{
debinfo("NewBranch\n");

  float dist, ftmp;
  int i, j, k, i2, j2, fathertrianglenum;
  branch *newbranch;
  vertex3f vertextmp;
  vector3f vectortmp;
  triangle3f triangletmp;
  mesh2 *pmeshtmp, *punionmesh;

  if(0 > father->offspringnum )
  {
    vertextmp[0] = 0.0;
    vertextmp[1] = 0.0;
    vertextmp[2] = 0.0;
    paMatrixVertex3f (vertextmp, father->midpoint);

    vertextmp[2] = width;
    paMatrixVertex3f (vertextmp, father->basepoints[0]);

    for (i = 1; i < BRANCHBASEPOINTS; i++)
    {
      vertextmp[0] = basesin[i]*width;
      vertextmp[1] = 0.0;
      vertextmp[2] = basecos[i]*width;
      paMatrixVertex3f (vertextmp, father->basepoints[i]);
    }

    father->offspringnum = 0;
  }

  newbranch = (branch*) malloc(sizeof(branch));
  newbranch->offspringnum = 0;
  newbranch->offsprings = NULL;
  newbranch->psurface = NULL;
  newbranch->father = father;

  father->offspringnum++;
  father->offsprings = (branch**) realloc(father->offsprings,
     father->offspringnum * sizeof(branch*));

  father->offsprings[father->offspringnum - 1] = newbranch;

  paTranslatef(0.0, length, 0.0);

    /* count the endpoints of new branch */
  CountNewBranchBasePoints(newbranch, father, width);

    /* create new branch... */
  pmeshtmp = NewMesh();

  i = BRANCHBASEPOINTS-1;
  i2 = i;

  for (j = BRANCHBASEPOINTS-2,
   (j2 = (i2 == 0) ? BRANCHBASEPOINTS-1 : i2-1);
    i >= 0; i--,
    (j = (j == 0) ? BRANCHBASEPOINTS-1 : j-1),
    (i2 = (i2 == 0) ? BRANCHBASEPOINTS-1 : i2-1),
    (j2 = (j2 == 0) ? BRANCHBASEPOINTS-1 : j2-1) )
  {
    memcpy (triangletmp[0], newbranch->basepoints[i],
        sizeof(vertex3f));
    memcpy (triangletmp[1], father->basepoints[j2],
        sizeof(vertex3f));
    memcpy (triangletmp[2], father->basepoints[i2],
        sizeof(vertex3f));
    AddTriangleToMesh(pmeshtmp, triangletmp);

    memcpy (triangletmp[0], father->basepoints[j2],
        sizeof(vertex3f));
    memcpy (triangletmp[1], newbranch->basepoints[i],
        sizeof(vertex3f));
    memcpy (triangletmp[2], newbranch->basepoints[j],
        sizeof(vertex3f));
    AddTriangleToMesh(pmeshtmp, triangletmp);
  }

  if (NULL == father->psurface)
  {
    father->psurface = pmeshtmp;
  }
  else
  {
    /* ... produce the union of old branches and new branch */
    punionmesh = NewMesh();
    Bool(father->psurface, pmeshtmp, punionmesh, UNION);
    FreeMesh(father->psurface);
    FreeMesh(pmeshtmp);
    father->psurface = punionmesh;
  }

debinfo("NewBranch end\n");
  return newbranch;
}


/* adds a copy of 'leaf' scaled by 'size' to 'pleafmesh'
  the position of the leaf is between the base and the end of 'pbranch'
    at 'pos', on the side determined by 'angle' */
int NewLeaf(mesh2 *pleafmesh, mesh *leaf, branch *pbranch,
  float pos, float angle, float size)
{
  int i;
  branch *pfather;
  vertex3f place, axis;
  triangle3f triangletmp;
  matrix4x4f matrix;

    // now, paMatrix transforms (0, 0, 0) to pbranch->midpoint
  paPushMatrix();
    // axis of newbranch
  memcpy(axis, pbranch->midpoint, sizeof(vertex3f));
  vertexsub3f(axis, pbranch->father->midpoint);

    /* place on the pbranch->father->basepoint[0] - pbranch->basepoint[0]
       line */
  memcpy(place, pbranch->father->basepoints[0], sizeof(vertex3f));
  vertexsub3f(place, pbranch->basepoints[0]);
  vertexmul3f(place, 1.0 - pos);
  vertexadd3f(place, pbranch->basepoints[0]);
  vertexsub3f(place, pbranch->midpoint);

  paOuterLoadIdentity(matrix);
  paOuterTranslatef(matrix, pbranch->midpoint[0],
    pbranch->midpoint[1], pbranch->midpoint[2]);
  paOuterRotatef(matrix, angle, axis[0], axis[1], axis[2]);
  paOuterTranslatef(matrix, place[0], place[1], place[2]);
/*  paOuterTranslatef(matrix, -pbranch->midpoint[0],
    -pbranch->midpoint[1], -pbranch->midpoint[2]);
*/
/*  paTranslatef(place[0], place[1], place[2]);
  paRotatef(angle, axis[0], axis[1], axis[2]);
*/

  for(i = 0; i < leaf->trianglenum; i++)
  {
    paMatrixVector3f(leaf->triangles[i][0], triangletmp[0]);
    paMatrixVector3f(leaf->triangles[i][1], triangletmp[1]);
    paMatrixVector3f(leaf->triangles[i][2], triangletmp[2]);

    paOuterMatrixVertex3f(matrix, triangletmp[0], triangletmp[0]);
    paOuterMatrixVertex3f(matrix, triangletmp[1], triangletmp[1]);
    paOuterMatrixVertex3f(matrix, triangletmp[2], triangletmp[2]);

    AddTriangleToMesh(pleafmesh, triangletmp);
  }

  pfather = pbranch->father;
  paPopMatrix();
  return 0;
}


int CountNewBranchBasePoints(branch *newbranch, branch *father, float width)
{
  vertex3f vertextmp, vertextmp2;
  float i, j, k, l, m, n, x, y, z, theta, omega, tmp;
  float a, b, c, deg;
  int ind;

  vertextmp[0] = 0.0;
  vertextmp[1] = 0.0;
  vertextmp[2] = 0.0;
  paMatrixVertex3f (vertextmp, newbranch->midpoint);


  x = newbranch->midpoint[0] - father->midpoint[0];
  y = newbranch->midpoint[1] - father->midpoint[1];
  z = newbranch->midpoint[2] - father->midpoint[2];

  omega = x*newbranch->midpoint[0] +
    y*newbranch->midpoint[1] +
    z*newbranch->midpoint[2];

  i = father->basepoints[0][0];
  j = father->basepoints[0][1];
  k = father->basepoints[0][2];

  l = (i + x);
  m = (j + y);
  n = (k + z);
  tmp =  -x*l - y*m - z*n;

  theta = (omega + tmp) / (x*i + y*j + z*k + tmp);
  a = i*theta + l*(1-theta) - newbranch->midpoint[0];
  b = j*theta + m*(1-theta) - newbranch->midpoint[1];
  c = k*theta + n*(1-theta) - newbranch->midpoint[2];

//@
/*  vertextmp[0] = father->basepoints[0] + x;
  vertextmp[0] = father->basepoints[0] + y;
  vertextmp[0] = father->basepoints[0] + z; */
/*  memcpy(vertextmp, father->basepoints[0], sizeof(vertex3f));
  vertexsub3f(vertextmp, father->midpoint);
*/    // vertextmp is now as long as the father's width
/*  a = z-y;
  b = x-z;
  c = y-x; */
//@
  tmp = width / sqrt(a*a + b*b + c*c);
  vertextmp[0] = a*tmp;
  vertextmp[1] = b*tmp;
  vertextmp[2] = c*tmp;
/* now, vertextmp is as long as the new width, and paralell to the line between
 father->midpoint - father->basepoints[0]
*/
//  vertextmp[2] = width;
  memcpy(newbranch->basepoints[0], vertextmp, sizeof(vertex3f));
  vertexadd3f(newbranch->basepoints[0], newbranch->midpoint);

  deg = 360 / BRANCHBASEPOINTS;

  paPushMatrix();
  paLoadIdentity();

  for(ind = 1; ind < BRANCHBASEPOINTS; ind++)
  {
    paRotatef(deg, x, y, z);
    paMatrixVertex3f(vertextmp, newbranch->basepoints[ind]);
//    memcpy(newbranch->basepoints[ind], vertextmp2, sizeof(vertex3f));
    vertexadd3f(newbranch->basepoints[ind], newbranch->midpoint);
  }

  paPopMatrix();
  return 0;
}


/* allocates space for 'vertexes' and 'normals' fills them and deletes the
  structure
  returns number of triangles */
int GetTrianglesAndDie(branch *root, vertex3f **vertexes, vector3f **normals)
{
debinfo("GetTrianglesAndDie\n");
  int trianglenum;
  vertex3f *vertextmp;
  vector3f *vectortmp;
  mesh2 *ptreemesh, *subdivmesh;

  ptreemesh = NewMesh();
  AddBranchMeshAndDieRec(root, ptreemesh);

  subdivmesh = NewMesh();
  SubDivide(ptreemesh, subdivmesh, MINLENGTH, W_DEFAULT);
  FreeMesh(ptreemesh);

  trianglenum = GetMeshTrianglesAndNormals3f(subdivmesh, vertexes, normals);
//  ptreemesh = NewMesh();
//  SubDivide(subdivmesh, ptreemesh, MINLENGTH, W_DEFAULT);
  FreeMesh(subdivmesh);

//  trianglenum = GetMeshTrianglesAndNormals3f(ptreemesh, vertexes, normals);
//  FreeMesh(ptreemesh);
debinfo("GetTrianglesAndDie end\n");
  return trianglenum;
}

/* adds the triangles of 'pbranch->psurface' to 'pmesh', and calls itself
 recursively on all offsprings of 'pbranch' */
int AddBranchMeshAndDieRec(branch *pbranch, mesh2 *pmesh)
{
  int i, j, imax;
  triangle3f triangletmp;

  if (NULL != pbranch->psurface)
  {
    imax = GetMeshTriangleNum(pbranch->psurface);
//printf("imax %d\n", imax);
    for (i = 0;i < imax; i++)
    {
      GetMeshTriangle3f(pbranch->psurface, i, triangletmp);
      AddTriangleToMesh(pmesh, triangletmp);
    }
    FreeMesh(pbranch->psurface);
  }
   // create cap at the end of branches
  else
  {
    for(i = 0, j = 1; i < BRANCHBASEPOINTS; i++, j = (j+1) % BRANCHBASEPOINTS)
    {
      memcpy(triangletmp[0], pbranch->basepoints[i], sizeof(vertex3f));
      memcpy(triangletmp[1], pbranch->basepoints[j], sizeof(vertex3f));
      memcpy(triangletmp[2], pbranch->midpoint, sizeof(vertex3f));

      AddTriangleToMesh(pmesh, triangletmp);
    }
  }

  if (pbranch->offspringnum > 0)
  {
    for(i = 0; i < pbranch->offspringnum; i++)
      AddBranchMeshAndDieRec(pbranch->offsprings[i], pmesh);

    free(pbranch->offsprings);
  }

  free(pbranch);
}



#ifdef WITHGL
#include <GL/glut.h>

#define SEGMENTS  6

  /* draws entire tree, given the turtle instruction string in 'tree' */
void OldDrawTree(char *tree)
{
  float param, width = 1;
  int ch, sign;
    // this will hold width of branch
  stack st;
  initstack(&st);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  while (*tree != 0)
  {
    sign = 1;
    switch(*tree)
    {
      case '[':
        glPushMatrix();
        push(&st, &width, sizeof(float));
        break;
      case ']':
        glPopMatrix();
        pop(&st, &width, sizeof(float));
        break;
      case 'F':
        tree+=2;
        getfloat(&tree, &param);
        DrawCylinder(width, param, SEGMENTS);
        glTranslatef(0.0, param, 0.0);
        break;
      case '!':
        tree+=2;
        getfloat(&tree, &param);
        width = param;
        break;
      case '-':
        sign = -1;
      case '+':
        tree+=2;
        getfloat(&tree, &param);
        param *= sign;
        glRotatef(param, 0.0, 1.0, 0.0);
        break;
      case '^':
        sign = -1;
      case '&':
        tree+=2;
        getfloat(&tree, &param);
        param *= sign;
        glRotatef(param, -1.0, 0.0, 0.0);
        break;
      case '/':
        sign = -1;
      case '\\':
        tree+=2;
        getfloat(&tree, &param);
        param *= sign;
        glRotatef(param, 0.0, 0.0, -1.0);
        break;
/*      case '|':
        glRotatef(180, 0.0, 1.0, 0.0);
        break;
*/
    }

    tree++;
  } // while (*tree != 0)

  glPopMatrix();
  freestack(&st);
} // OldDrawTree

  /* draws entire tree, given the turtle instruction string in 'tree' */
void DrawTree(char *tree)
{
  float param, width = 1;
  int ch, sign;
    // this will hold width of branch
  stack st;
  matrix4x4f *paMatrix;
  initstack(&st);

  glMatrixMode(GL_MODELVIEW);

  paLoadIdentity();

  while (*tree != 0)
  {
    sign = 1;
    switch(*tree)
    {
      case '[':
        paPushMatrix();
        push(&st, &width, sizeof(float));
        break;
      case ']':
        paPopMatrix();
        pop(&st, &width, sizeof(float));
        break;
      case 'F':
        tree+=2;
        getfloat(&tree, &param);
        glPushMatrix();
        paMatrix = paGetMatrixf();
        glMultMatrixf((float *) *paMatrix);

        DrawCylinder(width, param, SEGMENTS);
        glPopMatrix();

        paTranslatef(0.0, param, 0.0);
        break;
      case '!':
        tree+=2;
        getfloat(&tree, &param);
        width = param;
        break;
      case '-':
        sign = -1;
      case '+':
        tree+=2;
        getfloat(&tree, &param);
        param *= sign;
        paRotatef(param, 0.0, 1.0, 0.0);
        break;
      case '^':
        sign = -1;
      case '&':
        tree+=2;
        getfloat(&tree, &param);
        param *= sign;
        paRotatef(param, -1.0, 0.0, 0.0);
        break;
      case '/':
        sign = -1;
      case '\\':
        tree+=2;
        getfloat(&tree, &param);
        param *= sign;
        paRotatef(param, 0.0, 0.0, -1.0);
        break;
/*      case '|':
        glRotatef(180, 0.0, 1.0, 0.0);
        break;
*/
    }

    tree++;
  } // while (*tree != 0)

  freestack(&st);
} // DrawTree


 /* draws a cylinder standing on the xz plane  -- alters nothing just draws
quads */
void DrawCylinder(float radius, float height, int segments)
{
  float deg, x, z, curcos, cursin;
  int i;
  segments = segments < 3 ? 3 : segments;
  deg = 2 * M_PI / segments;

  glBegin(GL_QUAD_STRIP);

  glNormal3f(0.0, 0.0, 1.0);

  glVertex3f(0.0, 0.0, radius);
  glVertex3f(0.0, height, radius);

  for (i = 1; i < segments; i++)
  {
    cursin = sin(i*deg);
    curcos = cos(i*deg);
    x = cursin * radius;
    z = curcos * radius;

    glNormal3f(cursin, 0.0, curcos);
    glVertex3f(x, 0.0, z);
    glVertex3f(x, height, z);
  }

  glVertex3f(0.0, 0.0, radius);
  glVertex3f(0.0, height, radius);

  glEnd();

} // DrawCylinder

#endif // WITHGL


/* reads a float from 'str' and sets the pointer after it
  puts the float into num, returns 0 on success
  1 if no float can be read */
int getfloat(char **str, float *num)
{
  float value=0.0, dec;

#ifndef FLOATTEXT
  if (FLOATESCAPE == **str)
  {
    (*str) ++;
    memcpy(&value, (*str), sizeof(float));
    (*str) += sizeof(float);
  }
  else
#endif
  {
    if(!isdigit(**str))
      return 1;
    while(isdigit(**str))
    {
      value = 10*value + **str - '0';
      (*str)++;
    }
    if ('.' == **str)
    {
      (*str)++;
      dec = 0.1;
      while(isdigit(**str))
      {
        value += dec * (**str - '0');
        dec /= 10.0;
        (*str)++;
      }
    } //while
  } //else
  *num = value;
  return 0;
}


