/* turtle3d.h
graphically represent the turtle string
Prim Andras 2002
*/

#ifndef TURTLE3D_H
#define TURTLE3D_H

#include "coordgeom.h"
#include "lsystem.h"
#define BRANCHBASEPOINTS 6

#define DEFBRANCHLENGTH 0.2
#define DEFWIDTH        0.2
#define DEFLEAFPOS      0.5
#define DEFLEAFANGLE    0
#define DEFLEAFSIZE     1
#define DEFTURNANGLE    15

typedef struct _branch
{
  int offspringnum;
  struct _branch **offsprings, *father;
  mesh2 *psurface;
  vertex3f basepoints[BRANCHBASEPOINTS];
  vertex3f midpoint;
}branch;


/* reads a float from 'str' and sets the pointer after it
  puts the float into num, returns 0 on success
  1 if no float can be read */
int getfloat(char **str, float *num);


/* creates the mesh of a tree given with the string 'pl->sentence' in a form,
  that is easy to handle with GL_TRIANGLES
  returns number of triangles */
int CreateTreeMesh(plant *pl, vertex3f **vertexes, vector3f **normals);
/* creates the tree and adds the branches to ptreemesh and the leaves
 to pleafmesh */
int CreateLeafyTreeMesh(plant *pl, mesh2 *ptreemesh, mesh2 *pleafmesh);

/* creates a new branch at the endpoint of 'father' according to paMatrix, gives
himself to 'father->offsprings' and translates paMatrix to its endpoint.
returns a pointer to the new branch */
branch *NewBranch(branch* father, float width, float length);

/* adds a copy of 'leaf' scaled by 'size' to 'pleafmesh'
  the position of the leaf is between the base and the end of 'pbranch'
    at 'pos', on the side determined by 'angle' */
int NewLeaf(mesh2 *pleafmesh, mesh *leaf, branch *pbranch,
  float pos, float angle, float size);

/* counts the basepoints and midpoint of 'newbranch' in a way, so that
father.basepoint[i]-newbranch.basepoint[i] lines are paralell to
father.midpoint[i]-newbranch.midpoint */
int CountNewBranchBasePoints(branch *newbranch, branch *father, float width);

/* allocates space for 'vertexes' and 'normals' fills them and deletes the
  structure
  returns number of triangles */
int GetTrianglesAndDie(branch *root, vertex3f **vertexes, vector3f **normals);
/* adds the triangles of 'pbranch->psurface' to 'pmesh', and calls itself
 recursively on all offsprings of 'pbranch' and frees space allocated by
 'pbranch' */
int AddBranchMeshAndDieRec(branch *pbranch, mesh2 *pmesh);

#ifdef WITHGL
  /* draws entire tree, given the turtle instruction string in 'tree' */
void OldDrawTree(char *tree);
  /* draws entire tree, given the turtle instruction string in 'tree' */
void DrawTree(char *tree);

 /* draws a cylinder standing on the xz plane -- alters nothing just draws quads
*/
void DrawCylinder(float radius, float height, int segments);
#endif // WITHGL


#endif // first inclusion of this header
